import enum
import re
from datetime import datetime
from typing import Generic, TypeVar

from videoag_common.miscellaneous import *
from videoag_common.database import *
from .field import ApiSimpleColumnField, FieldContext

_O = TypeVar("_O", bound="ApiObject")


class ApiStringField(ApiSimpleColumnField[_O], Generic[_O]):
    
    def __init__(self,
                 max_length: int or None = None,
                 min_length: int = 0,
                 regex: str or None = None,
                 **kwargs):
        super().__init__(**kwargs, type_id="string")
        self.min_length = min_length
        self.max_length = max_length
        self._pattern = None if regex is None else re.compile(regex)
    
    def post_init(self, context: FieldContext):
        super().post_init(context)
        
        if not isinstance(self._column.type, sql.types.String):
            raise TypeError(f"SQL type for string field must be string, but is '{self._column.type}'")
        
        if self._column.type.length is not None:
            if self.max_length is not None and self.max_length > self._column.type.length:
                raise ValueError("Maximum length of field is higher than maximum length of sql type")
            if self.max_length is None:
                self.max_length = self._column.type.length
    
    def post_init_check(self, context: FieldContext):
        super().post_init_check(context)
        
        if self.max_length is None and self.include_in_config:
            raise ValueError("A maximum length must be specified for a string field which is included in the config. "
                             "Either with sql length or on api field")
    
    def _add_type_restrictions_to(self, config_entry: dict):
        super()._add_type_restrictions_to(config_entry)
        config_entry["min_length"] = self.min_length
        config_entry["max_length"] = self.max_length
        if self._pattern is not None:
            config_entry["pattern"] = self._pattern.pattern
    
    def _db_value_to_json(self, db_value) -> JsonTypes:
        return db_value
    
    def _json_value_to_db(self, session: SessionDb, object_db: _O, json_value: CJsonValue):
        value_db = json_value.as_string(min_length=self.min_length, max_length=self.max_length)
        if self._pattern is not None and not self._pattern.fullmatch(value_db):
            json_value.raise_error(f"Value must match pattern: {self._pattern.pattern}")
        
        return value_db


class ApiIntegerField(ApiSimpleColumnField[_O], Generic[_O]):
    
    def __init__(self,
                 min_value: int = MIN_VALUE_SINT32,
                 max_value: int = MAX_VALUE_SINT32,
                 **kwargs):
        super().__init__(**kwargs, type_id="int")
        self.min_value = min_value
        self.max_value = max_value
        
        if self.min_value > self.max_value:
            raise ValueError("Min value is greater than max value")
    
    def post_init_check(self, context: FieldContext):
        super().post_init_check(context)
        
        if not isinstance(self._column.type, sql.types.Integer):
            raise TypeError(f"SQL type for integer field must be integer, but is '{self._column.type}'")
    
    def _add_type_restrictions_to(self, config_entry: dict):
        super()._add_type_restrictions_to(config_entry)
        config_entry["min_value"] = self.min_value
        config_entry["max_value"] = self.max_value
    
    def _db_value_to_json(self, db_value) -> JsonTypes:
        return db_value
    
    def _json_value_to_db(self, session: SessionDb, object_db: _O, json_value: CJsonValue):
        return json_value.as_int(min_value=self.min_value, max_value=self.max_value)


class ApiBooleanField(ApiSimpleColumnField[_O], Generic[_O]):
    
    def __init__(self, **kwargs):
        super().__init__(**kwargs, type_id="boolean")
    
    def post_init_check(self, context: FieldContext):
        super().post_init_check(context)
        
        if self.may_be_none:
            raise ValueError("Boolean field may not be nullable")
        
        if not isinstance(self._column.type, sql.types.Boolean):
            raise TypeError(f"SQL type for boolean field must be boolean, but is '{self._column.type}'")
    
    def _db_value_to_json(self, db_value) -> JsonTypes:
        return db_value
    
    def _json_value_to_db(self, session: SessionDb, object_db: _O, json_value: CJsonValue):
        return json_value.as_bool()


class ApiDatetimeField(ApiSimpleColumnField[_O], Generic[_O]):
    
    def __init__(self, may_be_none: bool or None = None, **kwargs):
        super().__init__(**kwargs, type_id="datetime")
        self._may_be_none = may_be_none
    
    def post_init(self, context: FieldContext):
        super().post_init(context)
        
        if self._may_be_none is None:
            self._may_be_none = self._column.nullable
    
    def post_init_check(self, context: FieldContext):
        super().post_init_check(context)
        
        if not isinstance(self._column.type, sql.types.DateTime):
            raise TypeError(f"SQL type for datetime field must be datetime, but is '{self._column.type}'")
    
    @property
    def may_be_none(self) -> bool:
        return self._may_be_none
    
    def _db_value_to_json(self, db_value) -> JsonTypes:
        return db_value.strftime(DEFAULT_DATETIME_FORMAT)
    
    def _json_value_to_db(self, session: SessionDb, object_db: _O, json_value: CJsonValue):
        time_string = json_value.as_string(100)
        try:
            return datetime.strptime(time_string, DEFAULT_DATETIME_FORMAT)
        except ValueError:
            json_value.raise_error("Invalid datetime format")


class ApiEnumField(ApiSimpleColumnField[_O], Generic[_O]):
    
    def __init__(self, **kwargs):
        super().__init__(**kwargs, type_id="enum")
        self.enum_class = None
        self.str_enums: list[str] = []
        self.enums_by_id: dict[str, enum.Enum] = {}
    
    def post_init(self, context: FieldContext):
        super().post_init(context)
        
        if not isinstance(self._column.type, sql.types.Enum):
            raise TypeError(f"SQL type for enum field must be enum, but is '{self._column.type}'")
        
        self.enum_class = self._column.type.enum_class
        
        for enum in self.enum_class:
            if not isinstance(enum.value, str):
                raise ValueError(f"Enum '{self.enum_class}' for API field has non-string constant: '{enum.value}'")
            if enum.value in self.enums_by_id:
                raise ValueError(f"Enum '{self.enum_class}' for API field has duplicate constant: '{enum.value}'")
            self.enums_by_id[enum.value] = enum
        
        self.str_enums = list(self.enums_by_id.keys())
    
    def _add_type_restrictions_to(self, config_entry: dict):
        super()._add_type_restrictions_to(config_entry)
        config_entry["enums"] = self.str_enums
    
    def _db_value_to_json(self, db_value) -> JsonTypes:
        return db_value.value
    
    def _json_value_to_db(self, session: SessionDb, object_db: _O, json_value: CJsonValue):
        enum_id = json_value.as_string(1000)
        if enum_id not in self.enums_by_id:
            json_value.raise_error("Unknown enum value")
        return self.enums_by_id[enum_id]
