from abc import ABC, abstractmethod
from copy import copy
from dataclasses import dataclass
from typing import Generic, TypeVar, Callable, Any, Self

from videoag_common.miscellaneous import *
from videoag_common.database import *

_O = TypeVar("_O", bound="ApiObject")


@dataclass
class FieldContext:
    all_classes: dict[str, "ApiObjectClass"]
    api_class: "ApiObjectClass"
    own_member: Any
    variant_id: str or None


class ApiField(Generic[_O], ABC):
    
    def __init__(self, **kwargs):
        if len(kwargs) > 0:
            raise ValueError(f"Got unexpected argument(s): {kwargs.keys()}")
        self.variant_id = None
        
        self._api_class = None
        self.member_name = None
        self._own_member = None
    
    def set_member_name(self, member_name: str):
        self.member_name = member_name
    
    def pre_post_init_copy(self) -> Self:
        return copy(self)
    
    def post_init(self, context: FieldContext):
        self._api_class = context.api_class
        self.variant_id = context.variant_id
        self._own_member = context.own_member
    
    def post_init_check(self, context: FieldContext):
        pass
    
    def get_api_class(self) -> "ApiObjectClass":
        return self._api_class
    
    def __repr__(self):
        paras = []
        if isinstance(self, ApiConfigField) and self.include_in_config:
            paras.append(f"config_id={self.config_id}")
        
        if isinstance(self, ApiDataField) and self.include_in_data:
            paras.append(f"data_id={self.data_id}")
            
        return f"{type(self).__name__}[{', '.join(paras)}]"


class ApiConfigField(ApiField[_O], Generic[_O]):
    
    def __init__(self,
                 include_in_config: bool = False,
                 config_id: str or None = None,
                 config_only_at_creation: bool = False,
                 config_directly_modifiable: bool = False,
                 config_default_value: JsonTypes or tuple[None] or None = None,
                 **kwargs):
        super().__init__(**kwargs)
        self.include_in_config = include_in_config
        self.config_id = config_id
        self.config_only_at_creation = config_only_at_creation
        self.config_directly_modifiable = config_directly_modifiable
        self.config_default_value = config_default_value
        
        self.__description = None
    
    def post_init(self, context: FieldContext):
        super().post_init(context)
        
        if self.config_id is None:
            self.config_id = self.member_name
        
        if self.config_id is None:
            raise ValueError("No config id for field")
    
    def post_init_check(self, context: FieldContext):
        super().post_init_check(context)
        
        if self.config_directly_modifiable and not (
                isinstance(self, ApiDataField) and self.include_in_data
        ):
            raise ValueError("config_directly_modifiable can only be set when include_in_data is set")
    
    def config_get_load_options(self) -> list[ExecutableOption]:
        return []
    
    def config_get_description(self) -> JsonTypes or None:
        if not self.include_in_config:
            raise Exception("Config not enabled for field")
        if self.__description is not None:
            return self.__description
        
        self.__description = {
            "id": self.config_id,
            "type": self.config_get_type_id()
        }
        if self.config_default_value is not None:
            self.__description["default_value"] = (
                self.config_default_value if self.config_default_value != (None,) else None
            )
        self._add_type_restrictions_to(self.__description)
        return self.__description
    
    def config_get_value_with_description(self, object_db: _O) -> JsonTypes or None:
        if not self.include_in_config:
            raise Exception("Config not enabled for field")
        return {
            "description": self.config_get_description(),
            "value": self.config_get_value(object_db)
        }
    
    def _add_type_restrictions_to(self, config_entry: dict):
        """
        Added parameters need to be added to the api documentation (In routes/object_modifications.py)
        """
        pass
    
    @abstractmethod
    def config_get_type_id(self) -> str:
        pass
    
    @abstractmethod
    def config_get_value(self, object_db: _O) -> JsonTypes:
        pass  # pragma: no cover
    
    @abstractmethod
    def config_set_value(self,
                         session: SessionDb,
                         object_db: _O,
                         new_client_value: CJsonValue):
        pass  # pragma: no cover
    
    @abstractmethod
    def config_get_default(self) -> JsonTypes:
        """
        None indicates not default. (None, ) indicates a None default
        """
        pass  # pragma: no cover


class ApiDataField(ApiField[_O], Generic[_O]):
    
    def __init__(self,
                 include_in_data: bool = False,
                 data_id: str or None = None,
                 data_only_mod: bool = False,
                 data_if: Callable[[_O, dict], bool] or None = None,
                 data_notes: str = "",
                 **kwargs):
        super().__init__(**kwargs)
        self.include_in_data = include_in_data
        self.data_id = data_id
        self.data_only_mod = data_only_mod
        self._data_if = data_if
        self.data_notes = data_notes
    
    def post_init(self, context: FieldContext):
        super().post_init(context)
        
        if self.data_id is None:
            self.data_id = self.member_name
    
    def post_init_check(self, context: FieldContext):
        super().post_init_check(context)
        
        if self.data_id is None:
            raise ValueError("No data id for field")
    
    def data_is_optional(self) -> bool:
        return self._data_if is not None
    
    def add_to_data(self, object_db: _O, json: dict, args: Any):
        if not self.include_in_data:
            return
        if self._data_if and not self._data_if(object_db, args):
            return
        if self.data_only_mod and not args.is_mod:
            return
        json[self.data_id] = self.data_get_value(object_db, args)

    @abstractmethod
    def data_get_type_id(self) -> str:
        pass
    
    @abstractmethod
    def data_get_value(self, object_db: _O, args) -> JsonTypes:
        pass


class ApiAbstractColumnField(ApiConfigField[_O], ApiDataField[_O], Generic[_O]):
    
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        
        self._column: sql.Column or None = None
    
    def post_init_check(self, context: FieldContext):
        super().post_init_check(context)
    
    def config_get_value(self, object_db: _O) -> JsonTypes:
        db_value = getattr(object_db, self.member_name)
        if db_value is None:
            if not self.may_be_none and not self.ignore_db_null_for_never_none:
                raise TypeError(f"Got None value in db for field {self.config_id} but this field may not be None")
            return None
        return self._config_db_value_to_json(db_value)
    
    def config_set_value(self, session: SessionDb, object_db: _O, new_client_value: CJsonValue):
        if self.may_be_none and new_client_value.is_null():
            new_value_db = None
        else:
            new_value_db = self._config_json_value_to_db(session, object_db, new_client_value)
        setattr(object_db, self.member_name, new_value_db)
    
    def data_get_value(self, object_db: _O, args) -> JsonTypes:
        db_value = getattr(object_db, self.member_name)
        if self.may_be_none and db_value is None:
            return None
        return self._data_db_value_to_json(db_value, args)
    
    def config_get_default(self) -> JsonTypes or (None,):
        if self.config_default_value is not None:
            return self.config_default_value
        if self.may_be_none:
            return (None, )
        return None
    
    @property
    def may_be_none(self) -> bool:
        return False
    
    @property
    def ignore_db_null_for_never_none(self) -> bool:
        return False
    
    @abstractmethod
    def _data_db_value_to_json(self, db_value, args) -> JsonTypes:
        pass
    
    @abstractmethod
    def _config_db_value_to_json(self, db_value) -> JsonTypes:
        pass
    
    @abstractmethod
    def _config_json_value_to_db(self, session: SessionDb, object_db: _O, json_value: CJsonValue):
        pass


class ApiSimpleColumnField(ApiAbstractColumnField[_O], Generic[_O]):
    
    def __init__(self, type_id: str, unique: bool = False, **kwargs):
        super().__init__(**kwargs)
        self.type_id = type_id
        self.unique = unique
    
    # Note that unique is not in the specification on purpose. It's difficult for the client to check whether a given
    # value is unique. The easiest solution is to just send it to the server. If it's not unique the client will display
    # the error message (as it does for any error)
    
    def data_get_type_id(self) -> str:
        return self.type_id
    
    def config_get_type_id(self) -> str:
        return self.type_id
    
    def post_init(self, context: FieldContext):
        super().post_init(context)
        
        if not isinstance(context.own_member, orm.InstrumentedAttribute):
            raise TypeError(f"Member for column fields must be of type '{orm.InstrumentedAttribute}'")
        if not isinstance(context.own_member.prop, orm.ColumnProperty):
            raise TypeError(f"Property of member for column fields must be of type '{orm.ColumnProperty}'")
        
        if len(context.own_member.prop.columns) != 1:
            raise ValueError("Column property doesn't have exactly one column. This is not supported")
        
        column = context.own_member.prop.columns[0]
        if column.foreign_keys:
            print(f"Warning: Column '{column}' mapped in the API has foreign keys. For proper checks and "
                  f"errors, add an API mapping to an orm relationship instead.")
        
        self._column = column
        if self.config_default_value is None and self._column.default is not None:
            self.config_default_value = self._db_value_to_json(self._column.default.arg)
    
    def _config_json_value_to_db(self, session: SessionDb, object_db: _O, json_value: CJsonValue):
        db_value = self._json_value_to_db(session, object_db, json_value)
        
        if self.unique:
            query = object_db.basic_select().where(self._column == db_value)
            assert isinstance(query, sql.Select)
            if object_db.id is not None:
                # noinspection PyTypeChecker
                query = query.where(type(object_db).id != object_db.id)
            count = session.scalar(
                query.with_only_columns(sql.func.count())
            )
            if count > 0:
                raise ApiClientException(ERROR_OBJECT_ERROR(f"Value '{truncate_string(str(self._db_value_to_json(db_value)))}'"
                                                            f" for {self.config_id} is not unique"))
        return db_value
    
    def _config_db_value_to_json(self, db_value) -> JsonTypes:
        return self._db_value_to_json(db_value)
    
    def _data_db_value_to_json(self, db_value, args) -> JsonTypes:
        return self._db_value_to_json(db_value)
    
    @abstractmethod
    def _db_value_to_json(self, db_value) -> JsonTypes:
        pass
    
    @abstractmethod
    def _json_value_to_db(self, session: SessionDb, object_db: _O, json_value: CJsonValue):
        pass
