from abc import ABC
from typing import Generic, TypeVar

from videoag_common.database import *
from videoag_common.miscellaneous import *

from ..util import get_relationship_foreign_class
from .field import ApiAbstractColumnField, FieldContext

_O = TypeVar("_O", bound="ApiObject")


class ApiAbstractRelationshipField(ApiAbstractColumnField[_O], Generic[_O], ABC):
    
    def __init__(self,
                 data_foreign_in_context: bool = False,
                 **kwargs):
        super().__init__(**kwargs)
        self.data_foreign_in_context = data_foreign_in_context

        self._foreign_class = None
        self._relationship: orm.Relationship or None = None
    
    def post_init(self, context: FieldContext):
        super().post_init(context)
        
        if not isinstance(context.own_member, orm.InstrumentedAttribute):
            raise TypeError(f"Member for relationship fields must be of type '{orm.InstrumentedAttribute}'")
        if not isinstance(context.own_member.prop, orm.Relationship):
            raise TypeError(f"Property of member for relationship fields must be of type '{orm.Relationship}'")
        
        self._relationship = context.own_member.prop
        self._foreign_class = get_relationship_foreign_class(
            context.all_classes,
            self._relationship
        )
    
    def post_init_check(self, context: FieldContext):
        super().post_init_check(context)
        
        if self._relationship.viewonly and self.include_in_config:
            raise ValueError(f"include_in_config is True and viewonly is True for relationship '{self._relationship}'")
    
    @property
    def ignore_db_null_for_never_none(self) -> bool:
        # Relationship might have extra join conditions, and then it is always possible to get null values. A constraint
        # is often not feasible. In this case we just return None instead of throwing an exception
        return True


class ApiMany2OneRelationshipField(ApiAbstractRelationshipField[_O], Generic[_O]):
    
    def __init__(self, may_be_none: bool or None = None, **kwargs):
        super().__init__(**kwargs)
        self._may_be_none = may_be_none
    
    def post_init(self, context: FieldContext):
        self.config_id = f"{self.member_name}_id"
        
        if self.data_foreign_in_context and self.data_id is None and self.member_name is not None:
            self.data_id = f"{self.member_name}_id"
        
        super().post_init(context)
        
        if self._may_be_none is None:
            self._may_be_none = list(self._relationship.local_columns)[0].nullable
    
    def post_init_check(self, context: FieldContext):
        super().post_init_check(context)
        
        if len(self._relationship.local_columns) != 1:
            raise ValueError(f"Relationship '{self._relationship}' has more than one local column: "
                             f"'{self._relationship.local_columns}'")
        
        if self._relationship.collection_class is not None:
            raise ValueError(f"Relationship '{self._relationship}' may reference multiple objects. Use a many2many "
                             f"relationship instead")
    
    def config_get_load_options(self) -> list[ExecutableOption]:
        return super().config_get_load_options() + [
            orm.joinedload(self._own_member)
        ]
    
    @property
    def may_be_none(self) -> bool:
        return self._may_be_none
    
    def config_get_type_id(self) -> str:
        return self._foreign_class.id
    
    def data_get_type_id(self) -> str:
        return "int" if self.data_foreign_in_context else self._foreign_class.id
    
    def _data_db_value_to_json(self, db_value, args) -> JsonTypes:
        return self._foreign_class.serialize_object_args(db_value, args, to_context=self.data_foreign_in_context)
    
    def _config_db_value_to_json(self, db_value) -> JsonTypes:
        return db_value.id
    
    def _config_json_value_to_db(self, session: SessionDb, object_db: _O, json_value: CJsonValue):
        id = json_value.as_sint32()
        obj = session.scalar(
            self._foreign_class.orm_class.basic_select()
            .where(self._foreign_class.sql_id_column == id)
        )
        if obj is None:
            json_value.raise_error("Unknown object")
        return obj


class ApiAny2ManyRelationshipField(ApiAbstractRelationshipField[_O], Generic[_O]):
    """
    One class for many2many and one2many relationships. These are the same because in both cases we have a list of objects
    """
    
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
    
    def post_init(self, context: FieldContext):
        self.config_id = f"{self.member_name}_ids"
        
        if self.data_foreign_in_context and self.data_id is None and self.member_name is not None:
            self.data_id = f"{self.member_name}_ids"

        super().post_init(context)
    
    def post_init_check(self, context: FieldContext):
        super().post_init_check(context)
        
        if self._relationship.collection_class is None:
            raise ValueError(f"Relationship '{self._relationship}' can not reference multiple objects. Use a many2one "
                             f"relationship instead")
        
        if self.may_be_none:
            raise ValueError("may_be_none may not be set for many2many relationship")

    def config_get_load_options(self) -> list[ExecutableOption]:
        return super().config_get_load_options() + [
            orm.selectinload(self._own_member)
        ]
    
    def config_get_type_id(self) -> str:
        return self._foreign_class.id + "[]"
    
    def data_get_type_id(self) -> str:
        return ("int" if self.data_foreign_in_context else self._foreign_class.id) + "[]"
    
    def _data_db_value_to_json(self, db_value, args) -> JsonTypes:
        return list(map(
            lambda obj: self._foreign_class.serialize_object_args(obj, args, to_context=self.data_foreign_in_context),
            db_value
        ))
    
    def _config_db_value_to_json(self, db_value) -> JsonTypes:
        return [obj.id for obj in db_value]
    
    def _config_json_value_to_db(self, session: SessionDb, object_db: _O, json_value: CJsonValue):
        ids = []
        for val in json_value.as_array():
            ids.append(val.as_sint32())
        
        objects = session.scalars(
            self._foreign_class.orm_class.basic_select()
            .where(self._foreign_class.sql_id_column.in_(ids))
        ).all()
        if len(objects) != len(ids):
            missing_ids = set(ids) - set(map(lambda obj: obj.id, objects))
            json_value.raise_error(f"Unknown objects with ids: '{missing_ids}'")
        return objects
