from abc import abstractmethod
from typing import TypeVar, Callable

from videoag_common.miscellaneous import *
from videoag_common.database import *
from .fields import *

_T = TypeVar("_T")
_O = TypeVar("_O")


_MappedT = TypeVar("_MappedT", bound=orm.Mapped)


def api_mapped(
        mapped: _MappedT,
        field: ApiField,
) -> _MappedT:
    if isinstance(mapped, orm.MappedColumn):
        setattr(mapped.column, "video_ag_api_field", field)
    elif isinstance(mapped, orm.Relationship):
        setattr(mapped, "video_ag_api_field", field)
    else:
        raise TypeError(f"Unsupported mapped object: '{type(mapped)}'. Only MappedColumn and Relationship are supported")
    return mapped


def api_include_in_data(
    **kwargs
):
    field = ApiDataMethodField(include_in_data=True, **kwargs)
    
    def decorator(func):
        setattr(func, "video_ag_api_field", field)
        return func
    
    return decorator


class NoSerializationContext(dict):
    """
    A dict which can be used as a context for serialize() to prevent any serialization of objects of a specific type.
    This dict pretends like all keys are already present. Actually retrieving or updating an item throws an error
    """
    
    def __setitem__(self, __key, __value):
        raise Exception("Not supported")
    
    def __getitem__(self, __key):
        raise Exception("Not supported")
    
    def __contains__(self, __key):
        return True


API_VALUE_OBJECT_CLASSES_BY_ID: dict[str, type["ApiValueObject"]] = {}
_disable_value_object_registration = False


class ApiValueObject:
    
    def __init_subclass__(cls):
        super().__init_subclass__()
        if hasattr(cls, "__api_id__"):
            if _disable_value_object_registration:
                raise ValueError("Cannot register new ApiValueObject class. Make sure they are imported when initializing"
                                 "the orm object classes")
            
            id = cls.__api_id__
            if id in API_VALUE_OBJECT_CLASSES_BY_ID:
                raise ValueError(f"Duplicate value object id '{id}'")
            API_VALUE_OBJECT_CLASSES_BY_ID[id] = cls
    
    @abstractmethod
    def serialize(self) -> JsonTypes:
        pass


class RelationshipLoadMarker:
    pass


class ApiObject:
    id: Mapped[int] = api_mapped(
        mapped_column(nullable=False, primary_key=True, autoincrement=True, sort_order=-1),
        ApiIntegerField(
            include_in_data=True
        )
    )
    
    @staticmethod
    def __init_field_list(clazz):
        
        if hasattr(clazz, "__api_fields__"):
            api_fields = list(clazz.__api_fields__)
            delattr(clazz, "__api_fields__")  # Prevent that this variable is inherited by other classes
        else:
            api_fields = []
        
        setattr(clazz, "__all_class_api_fields__", api_fields)
        for name, member in clazz.__dict__.items():
            if isinstance(member, orm.MappedColumn) and hasattr(member.column, "video_ag_api_field"):
                field = member.column.video_ag_api_field
                delattr(member.column, "video_ag_api_field")
            elif hasattr(member, "video_ag_api_field"):
                field = member.video_ag_api_field
                delattr(member, "video_ag_api_field")
            else:
                continue
            field.set_member_name(name)
            api_fields.append(field)
    
    def __init_subclass__(cls, **kwargs):
        # We get the field data here, before calling init_subclass of the orm because the orm changes all the
        # attributes, and then we can't find our custom attributes
        
        # Also init list for own class if that hasn't happened yet
        if not hasattr(ApiObject, "__all_class_api_fields__"):
            ApiObject.__init_field_list(ApiObject)
        
        ApiObject.__init_field_list(cls)
        super().__init_subclass__(**kwargs)
    
    @hybrid_method
    def is_active(self, **kwargs):
        """
        Select conditions to get all objects which are still meaningful (e.g. everything which is not deleted)
        
        kwargs should all have default values and are only intended for optimization (e.g. prevent checking from both
        directions of a relationship)
        """
        if isinstance(self, type):
            return sql.True_()
        else:
            return True
    
    @hybrid_method
    def has_access(self, is_mod: bool, no_is_active_conditions: bool = False, **kwargs):
        """
        Select conditions to get all objects with some access restrictions. If is_mod=True, this must be effectively
        the same as is_active()
        """
        if no_is_active_conditions:
            if isinstance(self, type):
                return sql.True_()
            else:
                return True
        return self.is_active(**kwargs)
    
    @classmethod
    def basic_select(cls):
        return sql.select(cls).where(cls.is_active())
    
    @classmethod
    def select(cls, is_mod: bool, to_load: list[RelationshipLoadMarker], **kwargs):
        return (cls.basic_select()
                .where(cls.has_access(is_mod, no_is_active_conditions=True, **kwargs))
                .options(*cls.load_options(is_mod, to_load, **kwargs))
                )
    
    @classmethod
    def load_options(cls, is_mod: bool, to_load: list[RelationshipLoadMarker], **kwargs) -> list[ExecutableOption]:
        return []
    
    def serialize(self, **kwargs):
        return self.__api_class__.serialize(self, **kwargs)
    
    def serialize_to_context(self, **kwargs):
        return self.__api_class__.serialize_to_context(self, **kwargs)
    
    def __str__(self):
        return f"{type(self).__name__}[id={self.id}]"


ApiValueTypes = ApiObject or ApiValueObject or str or int or bool or None or list["ApiValueType"]


def api_get_value_serializer_function(
        all_classes: dict[str, "ApiObjectClass"],
        type_id: str,
        foreign_in_context: bool) -> tuple[str, Callable[[ApiValueTypes, ArgAttributeObject], JsonTypes]]:
    full_type_id = type_id
    is_array = False
    if type_id.endswith("[]"):
        is_array = True
        type_id = type_id[:len(type_id) - 2]
    
    _disable_value_object_registration = True
    
    if type_id in API_VALUE_OBJECT_CLASSES_BY_ID:
        if foreign_in_context:
            raise ValueError("foreign_in_context is set but this is not an object with an id")
        
        if type_id in all_classes:
            raise ValueError(f"Duplicate id '{type}' in API classes and value classes")
        value_class = API_VALUE_OBJECT_CLASSES_BY_ID[type_id]
        
        def serializer(val: ApiValueTypes, args):
            if val is None:
                return val
            if not isinstance(val, value_class):
                raise TypeError(f"Got object of type '{type(val)}' but it was previously declared to be a '{value_class}'")
            return val.serialize()
    elif type_id in all_classes:
        object_class = all_classes[type_id]
        
        if foreign_in_context:
            type_id = "int"
        
        def serializer(val: ApiValueTypes, args):
            if val is None:
                return val
            if not isinstance(val, object_class.orm_class):
                raise TypeError(f"Got object of type '{type(val)}' but it was previously declared to be a '{object_class.orm_class}'")
            return object_class.serialize_object_args(val, args, to_context=foreign_in_context)
    elif type_id in ("string", "integer", "boolean"):
        if foreign_in_context:
            raise ValueError("foreign_in_context is set but this is not an object with an id")
        
        def serializer(val: ApiValueTypes, args):
            return val
    else:
        raise ValueError(f"Unknown type '{full_type_id}' (For ApiValueObject make sure that the __api_id__ attribute is set)")
    
    if not is_array:
        return type_id, serializer
    
    def serializer_list(val_list: list[ApiValueTypes], args):
        if val_list is None:
            return val_list
        return list(map(lambda v: serializer(v, args), val_list))
    return f"{type_id}[]", serializer_list


class DeletableApiObject(ApiObject):
    deleted: Mapped[bool] = mapped_column(nullable=False, default=False)
    
    @hybrid_method
    def is_active(self, **kwargs):
        return super().is_active(**kwargs) & ~self.deleted


class VisibilityApiObject(ApiObject):
    visible: Mapped[bool] = api_mapped(
        mapped_column(sql.types.Boolean(), nullable=False, default=False),
        ApiBooleanField(
            include_in_config=True, config_directly_modifiable=True,
            include_in_data=True, data_only_mod=True
        )
    )
    
    @hybrid_method
    def has_access(self, is_mod: bool, visibility_check: bool = True, **kwargs):
        cond = super().has_access(is_mod, **kwargs)
        if not is_mod and visibility_check:
            cond &= self.visible
        return cond
