# Imports are used by files in objects/

from enum import Enum
from typing import Any

import sqlalchemy
from sqlalchemy import types, orm
from sqlalchemy.ext.hybrid import hybrid_method
from sqlalchemy.orm import DeclarativeBase, relationship
from sqlalchemy.sql.elements import AsBoolean

from videoag_common.miscellaneous import *

STRING_COLLATION = "und-x-icu"


def create_enum_type(enum_class: type[Enum]):
    return types.Enum(
        enum_class,
        values_callable=lambda x: [e.value for e in x],
        name=alphanum_camel_case_to_snake_case(enum_class.__name__)
    )


class Base(DeclarativeBase):
    
    def __init_subclass__(cls, **kw: Any) -> None:
        if not hasattr(cls, "__tablename__"):
            cls.__tablename__ = alphanum_camel_case_to_snake_case(cls.__name__)
        super().__init_subclass__(**kw)
    
    @hybrid_method
    def hybrid_rel(self, relationship, attribute: str, *args, **kwargs):
        """
        Helper method for hybrid methods. Can be used to specify conditions for relationships.
        
        Example: Consider a class 'Lecture' with an attribute 'course' and a class Course with an attribute 'deleted'.
        When you want to test whether the lecture's course is deleted with a hybrid method (e.g. in Python and SQL), you
        would need the expression 'self.course.deleted' for Python and 'self.course.has(Course.deleted)' for SQL. This
        can be replaced with the single expression 'self.hybrid_rel(self.course, "deleted")'
        
        The attribute may also be a method. In that case, all other arguments to this method, are given to that method
        """
        if isinstance(self, type):
            if not isinstance(relationship, orm.InstrumentedAttribute):
                raise TypeError("relationship must be orm attribute of self")
            if not isinstance(relationship.prop, orm.Relationship):
                raise TypeError("relationship must be a relationship")
            
            rel_class = relationship.prop.mapper.class_
            val = getattr(rel_class, attribute)
        else:
            val = getattr(relationship, attribute)
        
        if callable(val):
            val = val(*args, **kwargs)
        elif len(args) > 0 or len(kwargs) > 0:
            raise ValueError("Cannot give arguments for non-function attribute")
        
        if isinstance(self, type):
            # Small optimization to prevent SQLAlchemy from generating things like
            # EXISTS (SELECT 1 FROM course WHERE course.id = lecture.course_id AND true)
            if isinstance(val, (sqlalchemy.True_, sqlalchemy.False_)):
                return val
            if isinstance(val, AsBoolean) and isinstance(val.element, (sqlalchemy.True_, sqlalchemy.False_)):
                return val
            
            return relationship.has(val)
        else:
            return val
    
    @hybrid_method
    def hybrid_is_none(self, expression):
        if isinstance(self, type):
            return expression.is_(None)
        else:
            return expression is None
