import os
import sys
from enum import StrEnum
from typing import Callable, TypeVar, TypeVarTuple, Sequence

import sqlalchemy
from sqlalchemy import create_engine, URL
from sqlalchemy.orm import Session as SessionDb

from videoag_common.miscellaneous import *
from .base import Base
from .drift_detector import check_for_drift_and_migrate


_T = TypeVar("_T")
_A = TypeVarTuple("_A")


class TransactionIsolationLevel(StrEnum):
    READ_UNCOMMITTED = "READ UNCOMMITTED"
    READ_COMMITTED = "READ COMMITTED"
    REPEATABLE_READ = "REPEATABLE READ"
    SERIALIZABLE = "SERIALIZABLE"


class TransactionConflictError(Exception):
    pass


class Database:
    
    def __init__(self, config: dict):
        engine_type = config["engine"]
        if engine_type != "postgres":
            raise ValueError("The only supported engine is 'postgres'")
        
        engine_config = config[engine_type]
        host = dict_get_check_type(engine_config, "host", str)
        port = dict_get_check_type(engine_config, "port", int)
        user = dict_get_check_type(engine_config, "user", str)
        password = dict_get_check_type(engine_config, "password", str)
        database = dict_get_check_type(engine_config, "database", str)
        connect_args = engine_config.get("connect_args", {})
        
        check_drift = dict_get_check_type(engine_config, "check_drift", bool, True)
        auto_migration = dict_get_check_type(engine_config, "auto_migration", bool, False)
        skip_drift_check_when_no_connection = dict_get_check_type(engine_config, "skip_drift_check_when_no_connection", bool, False)
        if os.environ.get("API_SKIP_DRIFT_CHECK_WHEN_NO_CONNECTION", "false").lower() == "true":
            skip_drift_check_when_no_connection = True
        
        self._max_read_attempts = dict_get_check_type(engine_config, "max_read_attempts", int, 2)
        self._max_write_attempts = dict_get_check_type(engine_config, "max_write_attempts", int, 2)
        if self._max_read_attempts < 1 or self._max_write_attempts < 1:
            raise ValueError("Maximum read/write attempts must be >= 1")
        
        assert isinstance(user, str)
        assert isinstance(password, str)
        
        self._engine = create_engine(URL.create(
            "postgresql+psycopg",
            username=user,
            password=password,
            host=host,
            port=port,
            database=database
        ), connect_args=connect_args)
        
        self._read_engines_by_level = {}
        self._write_engines_by_level = {}
        for isolation_level in TransactionIsolationLevel:
            self._read_engines_by_level[isolation_level] = self._engine.execution_options(
                isolation_level=isolation_level.value,
                postgresql_readonly=True
            )
            self._write_engines_by_level[isolation_level] = self._engine.execution_options(
                isolation_level=isolation_level.value,
                postgresql_readonly=False
            )
        
        if check_drift:
            self._startup_check_drift(auto_migration, skip_drift_check_when_no_connection)
        
        if config.get("log_all_statements", False):
            import logging
            logger = logging.getLogger('sqlalchemy.engine')
            handler = logging.StreamHandler(sys.stdout)
            logger.addHandler(handler)
            logger.setLevel(logging.INFO)
    
    def _startup_check_drift(self, auto_migration: bool, skip_drift_check_when_no_connection: bool):
        if skip_drift_check_when_no_connection:
            session = SessionDb(bind=self._write_engines_by_level[TransactionIsolationLevel.REPEATABLE_READ])
            try:
                with session.begin():
                    session.execute(sqlachemy.text("SELECT 1"))
            except Exception as e:
                print(f"Unable to connect with database. Skipping drift check because "
                      f"skip_drift_check_when_no_connection is set. Exception: {e}")
                return
        
        # Ensure objects are loaded
        import videoag_common.objects
        
        drifted = not check_for_drift_and_migrate(Base.metadata, self._engine, auto_migration)
        if drifted:
            raise Exception("Database schema has drifted!")
    
    def query_all_and_expunge(self, stmt: sqlalchemy.Select[_T]) -> Sequence[_T]:
        def _trans(session: SessionDb):
            res = session.scalars(stmt).all()
            session.expunge_all()
            session.rollback()
            return res
        return self.execute_read_transaction(_trans)
    
    def query_one_and_expunge(self, stmt: sqlalchemy.Select[_T]) -> _T:
        def _trans(session: SessionDb):
            res = session.scalars(stmt).one()
            session.expunge_all()
            session.rollback()
            return res
        return self.execute_read_transaction(_trans)
    
    def query_one_or_none_and_expunge(self, stmt: sqlalchemy.Select[_T]) -> _T or None:
        def _trans(session: SessionDb):
            res = session.scalars(stmt).one_or_none()
            session.expunge_all()
            session.rollback()
            return res
        return self.execute_read_transaction(_trans)
    
    def execute_read_transaction(self,
                                 function: Callable[[SessionDb, *_A], _T],
                                 *args: *_A,
                                 isolation_level: TransactionIsolationLevel = TransactionIsolationLevel.REPEATABLE_READ
                                 ) -> _T:
        """
        Executes a read transaction with the given function. The function may be called multiple times if the read
        transaction fails due to conflicts with other transaction.

        Note that the specified isolation level is only a minimum. A higher isolation level may be used for the transaction.

        May raise :class:`NoAvailableConnectionError`
        """
        return self._execute_transaction(False, function, *args, isolation_level=isolation_level)

    def execute_write_transaction_and_commit(self,
                                             function: Callable[[SessionDb, *_A], _T],
                                             *args: *_A,
                                             isolation_level: TransactionIsolationLevel = TransactionIsolationLevel.SERIALIZABLE
                                             ) -> _T:
        """ Does the same as execute_write_transaction but commits the session automatically """
        def _trans(session: SessionDb, *trans_args) -> _T:
            res = function(session, *trans_args)
            session.commit()
            return res
        return self.execute_write_transaction(_trans, *args, isolation_level=isolation_level)
    
    def execute_write_transaction(self,
                                  function: Callable[[SessionDb, *_A], _T],
                                  *args: *_A,
                                  isolation_level: TransactionIsolationLevel = TransactionIsolationLevel.SERIALIZABLE
                                  ) -> _T:
        """
        Executes a read-write transaction with the given function. The function may be called multiple times if the write
        transaction fails due to conflicts with other transaction. The function is never called again if the transaction
        committed successfully

        Note that the specified isolation level is only a minimum. A higher isolation level may be used for the transaction.

        May raise :class:`NoAvailableConnectionError`
        """
        return self._execute_transaction(True, function, *args, isolation_level=isolation_level)
    
    def _execute_transaction(self,
                             writeable: bool,
                             function: Callable[[SessionDb, *_A], _T],
                             *args: *_A,
                             isolation_level: TransactionIsolationLevel = TransactionIsolationLevel.SERIALIZABLE
                             ) -> _T:
        attempts = 0
        while True:
            attempts += 1
            transaction = None
            try:
                engine = (self._write_engines_by_level if writeable else self._read_engines_by_level)[isolation_level]
                session = SessionDb(bind=engine)
                transaction = session.begin()
                result = function(session, *args)
                if transaction.is_active:
                    msg = ("Rolling open transaction back after execution but there was no exception (Always "
                           "commit/rollback the transaction explicitly)")
                    if writeable:
                        raise Exception(msg)
                    # noinspection PyBroadException
                    try:
                        print("Warning: " + msg)
                        transaction.rollback()
                    except Exception:
                        pass
                return result
            except Exception as e:
                is_conflict = Database._is_transaction_conflict_exception(e)
                
                if transaction is not None and transaction.is_active:
                    if not is_conflict:
                        self._on_transaction_aborted_by_user(writeable)
                    # noinspection PyBroadException
                    try:
                        transaction.rollback()
                    except:
                        pass
                
                if not is_conflict:
                    raise e

                self._on_transaction_conflict(writeable)
                if attempts >= (self._max_write_attempts if writeable else self._max_read_attempts):
                    self._on_transaction_aborted_after_repeated_conflict(writeable)
                    raise TransactionConflictError(e)
                continue
                
        raise AssertionError("This should be unreachable code")
    
    @staticmethod
    def _is_transaction_conflict_exception(e: Exception) -> bool:
        return "could not serialize access" in str(e)
    
    def _on_transaction_started(self, writeable: bool):
        pass
    
    def _on_transaction_aborted_by_user(self, writeable: bool):
        pass
    
    def _on_transaction_conflict(self, writeable: bool):
        pass
    
    def _on_transaction_aborted_after_repeated_conflict(self, writable: bool):
        pass
