import re
from typing import Any
import math

from videoag_common.database import *
from videoag_common.miscellaneous import *

from .error import MediaProcessException
from .target import TargetProducer, SingleInputTargetProducer, SingleOutputTargetProducer

SOURCE_FILE_TAG_PATTERN = re.compile("[a-zA-Z0-9_]*")


class SourceFileTargetProducer(SingleOutputTargetProducer["SourceMedium"]):
    
    def __init__(self, tag: str, **kwargs):
        super().__init__(**kwargs)
        self.tag = tag
    
    @classmethod
    def get_type(cls) -> str:
        return "source_file"
    
    @classmethod
    def _get_args_from_json(cls, json: CJsonObject) -> dict[str, Any]:
        tag = json.get_string("tag", 100)
        if not SOURCE_FILE_TAG_PATTERN.fullmatch(tag):
            json.get("tag").raise_error(f"Invalid tag. Must match {SOURCE_FILE_TAG_PATTERN.pattern}")
        return super()._get_args_from_json(json) | {
            "tag": tag
        }
    
    def to_json(self) -> dict:
        return super().to_json() | {
            "tag": self.tag
        }

    # Can't import Lecture, TargetMedium due to circular dependency
    # noinspection PyUnresolvedReferences
    def calculate_current_input_hash(self,
                                     session: SessionDb,
                                     lecture: "Lecture",
                                     input_media_by_id: dict[str, "TargetMedium"]) -> [str, "SourceMedium"]:
        from videoag_common.objects import SourceMedium
        source_medium = session.scalar(
            SourceMedium.basic_select().where(
                SourceMedium.lecture_id == lecture.id,
                SourceMedium.sorter_error_message.is_(None),
                SourceMedium.tag == self.tag
            )
            .order_by(SourceMedium.update_time.desc())
            .limit(1)
        )
        if source_medium is None:
            return "0" * 64, None
        return source_medium.sha256, source_medium

    # Can't import Lecture, TargetMedium due to circular dependency
    # noinspection PyUnresolvedReferences
    def create_new_target_media_and_job_data(
            self,
            session: SessionDb,
            lecture: "Lecture",
            input_media_by_id: dict[str, "TargetMedium"],
            common_new_medium_kwargs: dict[str, Any],
            source_medium: "SourceMedium",
    ) -> tuple[list["TargetMedium"], tuple[str, JsonTypes] or None] or None:
        from videoag_common.objects import SourceMedium, TargetMedium, TargetMediumType
        if source_medium is None:
            return None
        if source_medium.file_metadata is None:
            raise MediaProcessException(f"Source {source_medium.id} has no metadata")
        if source_medium.file_metadata["type"] != "target_medium":
            raise MediaProcessException(f"Source {source_medium.id} does not have 'target_medium' metadata")
        
        medium_data = source_medium.file_metadata["target_medium_data"]
        kwargs = medium_data | common_new_medium_kwargs | {
            "process_target_id": self.output_id,
            "file_path": source_medium.file_path,
            "is_produced": True,
        }
        medium_type = TargetMediumType(kwargs["type"])
        kwargs.pop("type")
        target_medium = orm.class_mapper(TargetMedium).polymorphic_map[medium_type].class_(**kwargs)
        
        return [target_medium], None


class DownscaleVideoTarget(SingleInputTargetProducer):
    
    def __init__(self, target_vertical_resolution: int, **kwargs):
        super().__init__(**kwargs)
        self.target_vertical_resolution = target_vertical_resolution
    
    @classmethod
    def get_type(cls) -> str:
        return "downscale_video"
    
    @classmethod
    def _get_args_from_json(cls, json: CJsonObject) -> dict[str, Any]:
        target_vertical_resolution = json.get_int("target_vertical_resolution", 1, 10000)
        return super()._get_args_from_json(json) | {
            "target_vertical_resolution": target_vertical_resolution
        }
    
    def to_json(self) -> dict:
        return super().to_json() | {
            "target_vertical_resolution": self.target_vertical_resolution
        }
    
    # Can't import Lecture, TargetMedium due to circular dependency
    # noinspection PyUnresolvedReferences
    def create_new_target_media_and_job_data(
            self,
            session: SessionDb,
            lecture: "Lecture",
            input_media_by_id: dict[str, "TargetMedium"],
            common_new_medium_kwargs: dict[str, Any],
            intermediate: None,
    ) -> tuple[list["TargetMedium"], tuple[str, JsonTypes] or None] or None:
        # TODO
        pass


class SampleThumbnailTarget(SingleInputTargetProducer, SingleOutputTargetProducer):
    
    def __init__(self, timestamp_percent: int, **kwargs):
        super().__init__(**kwargs)
        self.timestamp_percent = timestamp_percent
    
    @classmethod
    def get_type(cls) -> str:
        return "sample_thumbnail"
    
    @classmethod
    def _get_args_from_json(cls, json: CJsonObject) -> dict[str, Any]:
        timestamp_percent = json.get_int("timestamp_percent", 0, 100)
        return super()._get_args_from_json(json) | {
            "timestamp_percent": timestamp_percent
        }
    
    def to_json(self) -> dict:
        return super().to_json() | {
            "timestamp_percent": self.timestamp_percent
        }
    
    # Can't import Lecture, TargetMedium due to circular dependency
    # noinspection PyUnresolvedReferences
    def create_new_target_media_and_job_data(
            self,
            session: SessionDb,
            lecture: "Lecture",
            input_media_by_id: dict[str, "TargetMedium"],
            common_new_medium_kwargs: dict[str, Any],
            intermediate: None,
    ) -> tuple[list["TargetMedium"], tuple[str, JsonTypes] or None] or None:
        from videoag_common.objects import ThumbnailTargetMedium, PlainVideoTargetMedium, TargetMediumType
        input_medium = input_media_by_id[self.source_id]
        output_medium = ThumbnailTargetMedium(**(
            common_new_medium_kwargs | {
                "process_target_id": self.output_id
            }
        ))
        session.add(output_medium)
        session.flush()  # Ensure id is present
        output_medium.file_path = output_medium.get_default_file_path_no_ending() + ".jpg"
        
        if not isinstance(input_medium, PlainVideoTargetMedium):
            raise MediaProcessException(f"Input {self.source_id} is not of type {TargetMediumType.PLAIN_VIDEO}")
        
        if self.timestamp_percent == 0:
            time_sec = 0
        else:
            time_sec = math.ceil(input_medium.duration_sec * self.timestamp_percent / 100)
        
        return [output_medium], (
            "sample_thumbnail",
            {
                "input_file": input_medium.file_path,
                "output_file": output_medium.file_path,
                "time_sec": time_sec
            }
        )


# TODO
class AddIntroOutroTargetProducer(TargetProducer):
    pass


# TODO
class LogoOverlayTargetProducer(TargetProducer):
    pass
