from collections.abc import Iterable

from videoag_common.miscellaneous import *
from videoag_common.database import *
from videoag_common.api_object import ApiSimpleColumnField, FieldContext
import videoag_common
from .target import TargetProducer, TARGET_ID_PATTERN, TARGET_ID_MAX_LENGTH

#
# Important: Json default values should be avoided in case the default value is changed and an old process relied on the default
#

_PERMANENT_MEDIA_DIR_NAME = videoag_common.config["PERMANENT_MEDIA_DIR"]
if _PERMANENT_MEDIA_DIR_NAME.endswith("/"):
    raise ValueError("PERMANENT_MEDIA_DIR must NOT end with /")


# noinspection PyUnresolvedReferences
def get_permanent_lecture_dir(lecture: "Lecture") -> str:
    """
    Returns the db path for the directory of the given lecture. Does NOT end with a /
    """
    return (
        f"{_PERMANENT_MEDIA_DIR_NAME}/"
        f"course-{lecture.course.id}.{lecture.course.handle}/"
        f"lecture-{lecture.id}.{lecture.time.strftime("%y%m%d")}"
    )


class MediaProcess:
    
    def __init__(self, producers: list[TargetProducer], publish_target_ids: list[str], publish_wait_for_full_process: bool):
        super().__init__()
        self.producers = producers
        self.publish_target_ids = publish_target_ids
        self.publish_wait_for_full_process = publish_wait_for_full_process
        
        self.producer_by_output_id: dict[str, TargetProducer] = {}
        for prod in producers:
            for output_id in prod.get_target_output_ids():
                self.producer_by_output_id[output_id] = prod
        
        # Without any dependencies is first
        self._dependency_ordered: list[TargetProducer] = self._calculate_dependency_ordered()
        self._hash: str or None = None
    
    @property
    def producers_in_order(self) -> Iterable[TargetProducer]:
        return self._dependency_ordered
    
    def _calculate_dependency_ordered(self) -> list[TargetProducer]:
        # Without any dependencies is first
        ordered = []
        for producer in self.producers:
            self._add_with_dependencies_to_ordered(ordered, producer, set())
        return ordered
    
    def _add_with_dependencies_to_ordered(self, ordered: list[TargetProducer], target: TargetProducer, target_dependents: set[TargetProducer]):
        if target in ordered:
            return
        for dependency_id in target.get_target_input_ids():
            if dependency_id not in self.producer_by_output_id:
                raise ValueError(f"Unknown dependency target: {dependency_id}")
            
            dependency = self.producer_by_output_id[dependency_id]
            if dependency in target_dependents or dependency == target:
                raise ValueError(f"Cycle in dependencies. Cycle includes producers for targets {target.get_target_output_ids()} and {target.get_target_output_ids()}")
            self._add_with_dependencies_to_ordered(
                ordered,
                dependency,
                target_dependents.union([target])
            )
        ordered.append(target)
    
    @staticmethod
    def from_json(json: CJsonObject):
        try:
            target_ids: set[str] = set()
            producers = []
            for producer_json in json.get_array("producers"):
                producer = TargetProducer.from_json(producer_json.as_object())
                
                for output_id in producer.get_target_output_ids():
                    if not TARGET_ID_PATTERN.fullmatch(output_id):
                        producer_json.raise_error(f"Invalid output id '{output_id}'")
                    if output_id in target_ids:
                        producer_json.raise_error(f"Duplicate output target id '{output_id}'")
                    target_ids.add(output_id)
                
                for input_id in producer.get_target_input_ids():
                    if not TARGET_ID_PATTERN.fullmatch(input_id):
                        producer_json.raise_error(f"Invalid input id '{input_id}'")
                
                producers.append(producer)
            
            publish_target_ids = [
                id_json.as_string(TARGET_ID_MAX_LENGTH)
                for id_json in json.get_array("publish_target_ids")
            ]
            for target_id in publish_target_ids:
                if target_id not in target_ids:
                    json.raise_error(f"Unknown publish target id '{target_id}'")
            
            for producer in producers:
                for input_id in producer.get_target_input_ids():
                    if input_id not in target_ids:
                        json.raise_error(f"Unknown input target id '{input_id}' for producer of targets {producer.get_target_output_ids()}")
            
            return MediaProcess(producers, publish_target_ids, json.get_bool("publish_wait_for_full_process"))
        except Exception as e:
            raise Exception(f"With input:\n{json.get_unsafe_data()}") from e
    
    def to_json(self) -> dict:
        return {
            "producers": [
                producer.to_json()
                for producer in self.producers
            ],
            "publish_target_ids": self.publish_target_ids,
            "publish_wait_for_full_process": self.publish_wait_for_full_process
        }
    
    def sha256(self) -> str:
        if self._hash is None:
            self._hash = hash_json_sha256(self.to_json())
        return self._hash


class ApiMediaProcessField(ApiSimpleColumnField):
    
    def __init__(self, may_be_none: bool or None = None, **kwargs):
        super().__init__(**kwargs, type_id="media_process")
        self._may_be_none = may_be_none
    
    def post_init(self, context: FieldContext):
        super().post_init(context)
        
        if self._may_be_none is None:
            self._may_be_none = self._column.nullable
    
    def post_init_check(self, context: FieldContext):
        super().post_init_check(context)
        
        if not isinstance(self._column.type, sql.types.JSON):
            raise TypeError(f"SQL type for media process field must be json, but is '{self._column.type}'")
    
    @property
    def may_be_none(self) -> bool:
        return self._may_be_none
    
    def _db_value_to_json(self, db_value) -> JsonTypes:
        return db_value
    
    def _json_value_to_db(self, session: SessionDb, object_db, json_value: CJsonValue):
        # noinspection PyBroadException
        return MediaProcess.from_json(json_value.as_object()).to_json()
