import re
from abc import abstractmethod, ABC
from typing import Any, TypeVar, Generic

from videoag_common.miscellaneous import *
from videoag_common.database import *

TARGET_ID_PATTERN = re.compile("[a-zA-Z0-9_]{1,100}")
TARGET_ID_MAX_LENGTH = 100

_I = TypeVar("_I")


class TargetProducer(Generic[_I]):
    
    def __init__(self, **kwargs):
        if len(kwargs) > 0:
            raise ValueError(f"Got unknown kwargs: {kwargs.keys()}")
    
    def __str__(self):
        return f"{type(self).__name__}[output ids: {self.get_target_output_ids()}]"
    
    @staticmethod
    def from_json(json: CJsonObject) -> "TargetProducer":
        type = json.get_string("type", max_length=100)
        for clazz in recursive_flat_map_single(lambda c: c.__subclasses__(), TargetProducer):
            if clazz.get_type() != type:
                continue
            return clazz(**clazz._get_args_from_json(json))
    
    def to_json(self) -> dict:
        return {
            "type": self.get_type()
        }
    
    @classmethod
    @abstractmethod
    def get_type(cls) -> str:
        pass

    def get_target_input_ids(self) -> set[str]:
        return set()
    
    def get_target_output_ids(self) -> set[str]:
        return set()
    
    @classmethod
    def _get_args_from_json(cls, json: CJsonObject) -> dict[str, Any]:
        return {}
    
    # Can't import Lecture, TargetMedium due to circular dependency
    # noinspection PyUnresolvedReferences
    @abstractmethod
    def calculate_current_input_hash(self,
                                     session: SessionDb,
                                     lecture: "Lecture",
                                     input_media_by_id: dict[str, "TargetMedium"]) -> tuple[str, _I]:
        """
        This is intended to be only called from the process scheduler job
        
        When this is called, all dependency media are guaranteed to exist in the provided table
        """
        pass

    # Can't import Lecture, TargetMedium due to circular dependency
    # noinspection PyUnresolvedReferences
    @abstractmethod
    def create_new_target_media_and_job_data(self,
                                             session: SessionDb,
                                             lecture: "Lecture",
                                             input_media_by_id: dict[str, "TargetMedium"],
                                             common_new_medium_kwargs: dict[str, Any],
                                             intermediate: _I,
                                             ) -> tuple[list["TargetMedium"], tuple[str, JsonTypes] or None] or None:
        """
        This is intended to be only called from the process scheduler job.
        
        Creates a new target medium for every output of this producer and returns them. Additionally, returns the
        job's type and input_data to actually produce these target media. The job data may be None if no job needs to be
        scheduled (Because the work is so little, that the method did it itself).
        
        This method may also return None if the creation of the target medium is not in control of the backend (e.g.
        a user needs to upload a file). In this case it is ensured that such an action causes the media process
        scheduler job to be executed.
        
        The new target media's is_produced is True if the method returned no job (but not necessarily the other way).
        The new target medium may or may not have been added to session.
        
        :param session: The db session
        :param lecture: The lecture
        :param input_media_by_id: Contains all the media this producer depends on
        :param common_new_medium_kwargs: Contains all the values required by all target media except the
        process_target_id (e.g. lecture id, process sha, etc.)
        :param intermediate: The value returned by calculate_current_input_hash. May be used to prevent duplicate code
        """
        pass


class SingleInputTargetProducer(TargetProducer[None], ABC):
    
    def __init__(self, source_id: str, **kwargs):
        super().__init__(**kwargs)
        self.source_id = source_id
    
    @classmethod
    def _get_args_from_json(cls, json: CJsonObject) -> dict[str, Any]:
        source_id = json.get_string("source_id", TARGET_ID_MAX_LENGTH)
        return super()._get_args_from_json(json) | {
            "source_id": source_id
        }
    
    def to_json(self) -> dict:
        return super().to_json() | {
            "source_id": self.source_id
        }

    def get_target_input_ids(self) -> set[str]:
        return {self.source_id}

    # Can't import Lecture, TargetMedium due to circular dependency
    # noinspection PyUnresolvedReferences
    def calculate_current_input_hash(self,
                                     session: SessionDb,
                                     lecture: "Lecture",
                                     input_media_by_id: dict[str, "TargetMedium"]) -> [str, None]:
        return input_media_by_id[self.source_id].input_data_sha256, None


class SingleOutputTargetProducer(TargetProducer[_I], Generic[_I], ABC):
    
    def __init__(self, output_id: str, **kwargs):
        super().__init__(**kwargs)
        self.output_id = output_id
    
    @classmethod
    def _get_args_from_json(cls, json: CJsonObject) -> dict[str, Any]:
        output_id = json.get_string("output_id", TARGET_ID_MAX_LENGTH)
        return super()._get_args_from_json(json) | {
            "output_id": output_id
        }
    
    def to_json(self) -> dict:
        return super().to_json() | {
            "output_id": self.output_id
        }
    
    def get_target_output_ids(self) -> set[str]:
        return {self.output_id}
