import re
from typing import Iterable

from videoag_common.miscellaneous.util import JsonTypes


class CJsonException(Exception):
    
    def __init__(self, path: str, message: str):
        super().__init__(f"At {path}: {message}")
        self.path = path
        self.message = message


class CJsonValue:
    
    def __init__(self, data, path: str = "$"):
        self._data = data
        self._path = path
    
    def raise_error(self, message: str):
        raise CJsonException(self._path, message)
    
    def equals_json(self, json: JsonTypes):
        return self._data == json
    
    def is_null(self):
        return self._data is None
    
    def as_object(self) -> "CJsonObject":
        self.__check_value_type(dict, "object")
        return CJsonObject(self._data, self._path)
    
    def as_array(self) -> "CJsonArray":
        self.__check_value_type(list, "array")
        return CJsonArray(self._data, self._path)
    
    def as_bool(self) -> bool:
        self.__check_value_type(bool, "boolean")
        return self._data
    
    def as_sint32(self) -> int:
        return self.as_int(-2147483648, 2147483647)
    
    def as_int(self, min_value: int, max_value: int) -> int:
        self.__check_value_type(int, "int")
        if self._data < min_value:
            self.raise_error(f"Value must not be smaller than {min_value}")
        if self._data > max_value:
            self.raise_error(f"Value must not be greater than {max_value}")
        return self._data
    
    def as_string(self, max_length: int, min_length: int = None) -> str:
        self.__check_value_type(str, "string")
        if min_length is not None and len(self._data) < min_length:
            self.raise_error(f"String must not be shorter than {min_length}")
        if len(self._data) > max_length:
            self.raise_error(f"String must not be longer than {max_length}")
        return self._data
    
    def is_string(self):
        return isinstance(self._data, str)
    
    def get_unsafe_data(self) -> JsonTypes:
        return self._data
    
    def __check_value_type(self, expected_type: type, type_name: str):
        if self._data is None or not isinstance(self._data, expected_type):
            self.raise_error(f"Expected {type_name}")
        return self._data


class CJsonObject(CJsonValue):
    
    def __init__(self, data: {}, path: str = "$"):
        super().__init__(data, path)
    
    def has(self, key: str):
        return key in self._data
    
    def keys(self):
        return self._data.keys()
    
    def raise_if_present(self, key: str):
        if self.has(key):
            self.get(key).raise_error("This value should not be present")
    
    def get(self, key: str, optional: bool = False):
        if optional and not self.has(key):
            return None
        if key not in self._data:
            raise CJsonException(f"{self._path}.{key}", "Is missing")
        return CJsonValue(self._data[key], f"{self._path}.{key}")
    
    def get_object(self, key: str) -> "CJsonObject":
        return self.get(key).as_object()
    
    def get_array(self, key: str) -> "CJsonArray":
        return self.get(key).as_array()
    
    def get_bool(self, key: str, default: bool or None = None) -> bool:
        if default is not None and not self.has(key):
            return default
        return self.get(key).as_bool()
    
    def get_sint32(self, key: str, default: int or None = None) -> int:
        if default is not None and not self.has(key):
            return default
        return self.get(key).as_sint32()
    
    def get_int(self, key: str, min_value: int, max_value: int, optional: bool = False) -> int or None:
        if optional and not self.has(key):
            return None
        return self.get(key).as_int(min_value, max_value)
    
    def get_string(self, key: str, max_length: int, min_length: int = None, optional: bool = False) -> str or None:
        if optional and not self.has(key):
            return None
        return self.get(key).as_string(max_length, min_length)


class _ArrayIterator(Iterable[CJsonValue]):
    
    def __init__(self, array: "CJsonArray"):
        super().__init__()
        self._array = array
        self._i = 0
    
    def __next__(self):
        if self._i >= self._array.length():
            raise StopIteration
        ele = self._array.get(self._i)
        self._i += 1
        return ele
    
    def __iter__(self):
        return self


class CJsonArray(CJsonValue, Iterable[CJsonValue]):
    
    def __init__(self, data: [], path: str = "$"):
        super().__init__(data, path)
        
    def length(self) -> int:
        return len(self._data)
    
    def get(self, index: int):
        if index < 0:
            raise ValueError(f"Negative index provided: {index}")  # pragma: no cover
        if index >= self.length():
            raise CJsonException(f"{self._path}[{index}]", "Is missing")
        return CJsonValue(self._data[index], f"{self._path}[{index}]")
    
    def get_object(self, index: int) -> "CJsonObject":
        return self.get(index).as_object()
    
    def get_array(self, index: int) -> "CJsonArray":
        return self.get(index).as_array()
    
    def get_bool(self, index: int) -> bool:
        return self.get(index).as_bool()
    
    def get_sint32(self, index: int) -> int:
        return self.get(index).as_sint32()
    
    def get_int(self, index: int, min_value: int, max_value: int) -> int:
        return self.get(index).as_int(min_value, max_value)
    
    def get_string(self, index: int, max_length: int, min_length: int = None) -> str:
        return self.get(index).as_string(max_length, min_length)
    
    def __iter__(self):
        return _ArrayIterator(self)


_JSON_ARGUMENT_PATTERN = re.compile("\\$\\{([^{}]*?)}")


def replace_json_arguments(data: JsonTypes, argument_data: JsonTypes) -> JsonTypes:
    if isinstance(data, int) or isinstance(data, float) or isinstance(data, bool) or data is None:
        return data
    if isinstance(data, list):
        return [replace_json_arguments(e, argument_data) for e in data]
    if isinstance(data, dict):
        return {
            replace_json_arguments(key, argument_data): replace_json_arguments(value, argument_data)
            for key, value in data.items()
        }
    if not isinstance(data, str):
        raise TypeError(f"Unknown json type: {type(data)}")
    
    def _get_match_value(match, force_string: bool = True):
        path = match.group(1)
        val = argument_data
        for path_ele in path.split("."):
            if not isinstance(val, dict) or path_ele not in val:
                raise ValueError(f"Can't replace with path '{path}' because element '{path_ele}' is not present "
                                 f"(or parent has bad type) in argument data")
            val = val[path_ele]
        if force_string:
            if isinstance(val, str) or isinstance(val, int):
                val = str(val)
            else:
                raise ValueError(f"Can't replace with path '{path}' because result is not a str or int")
        return val
    
    full_match = _JSON_ARGUMENT_PATTERN.fullmatch(data)
    if full_match is not None:
        return _get_match_value(full_match, force_string=False)
    
    return _JSON_ARGUMENT_PATTERN.sub(_get_match_value, data)

