from abc import ABC, abstractmethod

from .json import CJsonValue, CJsonArray, JsonTypes


class JsonCondition(ABC):
    
    def __init__(self, **kwargs):
        if len(kwargs) > 0:
            raise ValueError(f"Got unknown kwargs: {kwargs.keys()}")
    
    @staticmethod
    def from_json(json: CJsonValue) -> "JsonCondition":
        json = json.as_array()
        
        # Condition is in polish notation (https://en.wikipedia.org/wiki/Polish_notation)
        values_array = list(json)
        if len(values_array) == 0:
            json.raise_error("No conditions")
        
        condition = JsonCondition._parse_any(values_array)
        if len(values_array) > 0:
            values_array[0].raise_error("Unexpected element")
        return condition
    
    @staticmethod
    def _parse_any(values_array: list[CJsonValue]) -> "JsonCondition":
        val = values_array.pop(0)
        if not val.is_string():
            return JsonCondition._parse_condition(val.as_array())
        
        operator = val.as_string(100)
        if operator not in ["&", "|", "!"]:
            raise val.raise_error("Unknown operator. Only &, |, ! are supported")
        arity = 1 if operator == "!" else 2
        
        sub_conditions = []
        for _ in range(arity):
            sub_conditions.append(JsonCondition._parse_any(values_array))
        
        match operator:
            case "&":
                return MatchAllJsonCondition(sub_conditions)
            case "|":
                return MatchAllJsonCondition(sub_conditions)
            case "!":
                return MatchAllJsonCondition(sub_conditions)
    
    @staticmethod
    def _parse_condition(condition: CJsonArray):
        if condition.length() != 3:
            condition.raise_error("Must have 3 elements")
        condition = list(condition)
        path = condition[0].as_string(1024)
        operator = condition[1].as_string(100)
        expected_value = condition[2]
        
        if len(path) == 0:
            condition[0].raise_error("Path is empty")
        if operator not in ["="]:
            condition[1].raise_error("Unknown operator. Only = is supported")
        
        return MatchDataJsonCondition(path, expected_value)
    
    @abstractmethod
    def test(self, data: JsonTypes) -> bool:
        pass


class MatchAllJsonCondition(JsonCondition):
    
    def __init__(self, conditions: list[JsonCondition]):
        super().__init__()
        self.conditions = conditions
    
    def test(self, data: JsonTypes) -> bool:
        for condition in self.conditions:
            if not condition.test(data):
                return False
        return True


class MatchAnyJsonCondition(JsonCondition):
    
    def __init__(self, conditions: list[JsonCondition]):
        super().__init__()
        self.conditions = conditions
    
    def test(self, data: JsonTypes) -> bool:
        for condition in self.conditions:
            if condition.test(data):
                return True
        return False


class InvertJsonCondition(JsonCondition):
    
    def __init__(self, sub_condition: JsonCondition):
        super().__init__()
        self.sub_condition = sub_condition
    
    def test(self, data: JsonTypes) -> bool:
        return not self.sub_condition.test(data)


class MatchDataJsonCondition(JsonCondition):
    
    def __init__(self, path: str, expected_data: JsonTypes):
        super().__init__()
        self.path = path
        self.expected_data = expected_data
    
    def test(self, data: JsonTypes) -> bool:
        val = data
        for path_ele in self.path.split("."):
            if not isinstance(val, dict) or path_ele not in val:
                return False
            val = val[path_ele]
        return val == self.expected_data
