from datetime import datetime
from enum import Enum

from videoag_common.database import *
from videoag_common.api_object import *
from .user import User


class ChangelogEntryType(Enum):
    MODIFICATION = "modification"
    DELETION_CHANGE = "deletion_change"
    CREATION = "creation"
    UNKNOWN = "unknown"


_CHANGELOG_ENTRY_TYPE_ENUM = create_enum_type(ChangelogEntryType)


LOAD_CHANGELOG_ENTRY_USER = RelationshipLoadMarker()


class ChangelogEntry(ApiObject, Base):
    __mapper_args__ = {
        "polymorphic_on": "type",
        "with_polymorphic": "*"  # Always load all attributes for all types
    }
    
    change_time: Mapped[datetime] = api_mapped(
        mapped_column(TIMESTAMP(), nullable=False, index=True, server_default=sql.text("CURRENT_TIMESTAMP")),
        ApiDatetimeField(
            include_in_data=True
        )
    )
    modifying_user_id: Mapped[int] = mapped_column(ForeignKey("user.id"), nullable=False, index=True)
    
    type: Mapped[ChangelogEntryType] = api_mapped(
        mapped_column(_CHANGELOG_ENTRY_TYPE_ENUM, nullable=False),
        ApiEnumField(
            include_in_data=True
        )
    )
    
    object_type: Mapped[str] = api_mapped(
        mapped_column(String(collation=STRING_COLLATION), nullable=False, index=True),
        ApiStringField(
            include_in_data=True, data_notes="Note that this object type may not exist anymore"
        )
    )
    object_id: Mapped[int] = api_mapped(
        mapped_column(nullable=False, index=True),
        ApiIntegerField(
            include_in_data=True
        )
    )
    
    modifying_user: Mapped[User] = api_mapped(
        relationship(
            primaryjoin=lambda: User.id == ChangelogEntry.modifying_user_id,
            lazy="raise_on_sql"
        ),
        ApiMany2OneRelationshipField(
            include_in_data=True, data_foreign_in_context=True
        )
    )
    
    @classmethod
    def load_options(cls, is_mod: bool, to_load: list[RelationshipLoadMarker], **kwargs) -> list[ExecutableOption]:
        options = super().load_options(is_mod, to_load, **kwargs)
        
        if LOAD_CHANGELOG_ENTRY_USER in to_load:
            options.append(orm.selectinload(ChangelogEntry.modifying_user))
        
        return options


class ChangelogModificationEntry(ChangelogEntry):
    __tablename__ = None  # Prevent our own base from adding a table name. This should be a single-table inheritance
    __mapper_args__ = {
        "polymorphic_identity": ChangelogEntryType.MODIFICATION
    }
    
    field_id: Mapped[str] = api_mapped(
        mapped_column(String(collation=STRING_COLLATION), nullable=True),
        ApiStringField(
            include_in_data=True, data_notes="Note that this field may not exist anymore"
        )
    )
    old_value: Mapped[dict] = api_mapped(
        mapped_column(sql.types.JSON(), nullable=True),
        ApiJsonField(
            include_in_data=True, data_if=lambda entry, args: entry.old_value is not None,
            data_notes="Note that this may contain arbitrary json due to legacy values. Not present if this was at creation"
        )
    )
    new_value: Mapped[dict] = api_mapped(
        mapped_column(sql.types.JSON(), nullable=True),
        ApiJsonField(
            include_in_data=True, data_notes="Note that this may contain arbitrary json due to legacy values"
        )
    )


class ChangelogDeletionChangeEntry(ChangelogEntry):
    __tablename__ = None
    __mapper_args__ = {
        "polymorphic_identity": ChangelogEntryType.DELETION_CHANGE
    }
    
    is_now_deleted: Mapped[bool] = api_mapped(
        mapped_column(nullable=False, default=False),
        ApiBooleanField(
            include_in_data=True
        )
    )


class ChangelogCreationEntry(ChangelogEntry):
    __tablename__ = None
    __mapper_args__ = {
        "polymorphic_identity": ChangelogEntryType.CREATION
    }

    parent_type: Mapped[str] = api_mapped(
        mapped_column(String(collation=STRING_COLLATION), nullable=True),
        ApiStringField(
            include_in_data=True
        )
    )
    parent_id: Mapped[int] = api_mapped(
        mapped_column(nullable=True),
        ApiIntegerField(
            include_in_data=True
        )
    )
    variant: Mapped[str] = api_mapped(
        mapped_column(String(collation=STRING_COLLATION), nullable=True),
        ApiStringField(
            include_in_data=True
        )
    )


class ChangelogUnknownEntry(ChangelogEntry):
    __tablename__ = None
    __mapper_args__ = {
        "polymorphic_identity": ChangelogEntryType.UNKNOWN
    }

    data: Mapped[dict] = api_mapped(
        mapped_column(sql.types.JSON, nullable=True),
        ApiJsonField(
            include_in_data=True
        )
    )
