from datetime import datetime

from videoag_common.database import *
from videoag_common.api_object import *
from videoag_common.media_process import *
from videoag_common.miscellaneous import *
from .view_permissions import EffectiveViewPermissions, ApiViewPermissionsObject, DEFAULT_VIEW_PERMISSIONS

LOAD_COURSE_LECTURES = RelationshipLoadMarker()
LOAD_LECTURE_COURSE = RelationshipLoadMarker()
LOAD_LECTURE_CHAPTERS = RelationshipLoadMarker()
LOAD_LECTURE_PUBLISH_MEDIA = RelationshipLoadMarker()
LOAD_LECTURE_TARGET_MEDIA = RelationshipLoadMarker()
LOAD_LECTURE_SOURCE_MEDIA = RelationshipLoadMarker()


class Chapter(DeletableApiObject, VisibilityApiObject, Base):
    __api_class__ = ApiObjectClass(
        parent_relationship_config_ids=["lecture"]
    )
    
    lecture_id: Mapped[int] = mapped_column(ForeignKey("lecture.id"), nullable=False, index=True)
    # TODO check duration
    start_time: Mapped[int] = api_mapped(
        mapped_column(nullable=False),
        ApiIntegerField(
            min_value=0,
            include_in_config=True, config_directly_modifiable=True,
            include_in_data=True, data_notes="In seconds, >= 0"
        )
    )
    name: Mapped[str] = api_mapped(
        mapped_column(Text(collation=STRING_COLLATION), nullable=False),
        ApiStringField(
            max_length=256,
            include_in_config=True, config_directly_modifiable=True,
            include_in_data=True
        )
    )
    
    lecture: Mapped["Lecture"] = relationship(
        primaryjoin=lambda: Lecture.id == Chapter.lecture_id,
        back_populates="chapters",
        lazy="raise_on_sql"
    )


responsible_table = sql.Table(
    "course_responsible",
    Base.metadata,
    sql.Column("course_id", ForeignKey("course.id"), primary_key=True),
    sql.Column("user_id", ForeignKey("user.id"), primary_key=True),
)


class Lecture(DeletableApiObject, VisibilityApiObject, ApiViewPermissionsObject, Base):
    __api_class__ = ApiObjectClass(
        parent_relationship_config_ids=["course"]
    )
    
    course_id: Mapped[int] = mapped_column(ForeignKey("course.id"), nullable=False, index=True)
    # We have indices on all the columns used in the search
    title: Mapped[str] = api_mapped(
        mapped_column(Text(collation=STRING_COLLATION), nullable=False, index=True, default=""),
        ApiStringField(
            min_length=1, max_length=256,
            include_in_config=True, config_directly_modifiable=True,
            include_in_data=True
        )
    )
    speaker: Mapped[str] = api_mapped(
        mapped_column(Text(collation=STRING_COLLATION), nullable=False, index=True, default=""),
        ApiStringField(
            max_length=256,
            include_in_config=True, config_directly_modifiable=True,
            include_in_data=True,
            data_notes="Name of the speaker"
        )
    )
    location: Mapped[str] = api_mapped(
        mapped_column(Text(collation=STRING_COLLATION), nullable=False, default=""),
        ApiStringField(
            max_length=256,
            include_in_config=True, config_directly_modifiable=True,
            include_in_data=True
        )
    )
    time: Mapped[datetime] = api_mapped(
        mapped_column(TIMESTAMP(), nullable=False, index=True),
        ApiDatetimeField(
            include_in_config=True, config_directly_modifiable=True,
            include_in_data=True
        )
    )
    duration: Mapped[int] = api_mapped(
        mapped_column(nullable=False, default=90),
        ApiIntegerField(
            min_value=0, max_value=7 * 24 * 60,
            include_in_config=True, config_directly_modifiable=True,
            include_in_data=True,
            data_notes="In minutes"
        )
    )
    description: Mapped[str] = api_mapped(
        mapped_column(Text(collation=STRING_COLLATION), nullable=False, index=True, default=""),
        ApiStringField(
            max_length=8192,
            include_in_config=True, config_directly_modifiable=True,
            include_in_data=True
        )
    )
    no_recording: Mapped[bool] = api_mapped(
        mapped_column(nullable=False, default=False),
        ApiBooleanField(
            include_in_config=True, config_directly_modifiable=True,
            include_in_data=True
        )
    )
    livestream_planned: Mapped[bool] = api_mapped(
        mapped_column(nullable=False, default=False),
        ApiBooleanField(
            include_in_config=True,
            include_in_data=True
        )
    )
    internal_comment: Mapped[str] = api_mapped(
        mapped_column(Text(collation=STRING_COLLATION), nullable=False, default=""),
        ApiStringField(
            max_length=8192,
            include_in_config=True, config_directly_modifiable=True,
            include_in_data=True, data_only_mod=True
        )
    )
    publish_time: Mapped[datetime] = api_mapped(
        mapped_column(TIMESTAMP(), nullable=True, index=True),
        ApiDatetimeField(
            include_in_data=True, data_only_mod=True
        )
    )
    media_process: Mapped[dict] = api_mapped(
        mapped_column(postgresql.JSONB, nullable=True),  # Null indicates, inherit from course
        ApiMediaProcessField(
            include_in_config=True
        )
    )
    disable_automatic_media_process_scheduler: Mapped[bool] = api_mapped(
        mapped_column(nullable=False, default=False),
        ApiBooleanField(
            include_in_config=True,
            # Course also has this field. Processor only works automatically if both fields are False
        )
    )
    current_media_process_error: Mapped[str] = mapped_column(nullable=True)
    
    @staticmethod
    def __lambda_primary_join_publish_media():
        from .medium import PublishMedium
        return sql.and_(
            PublishMedium.lecture_id == Lecture.id,
            PublishMedium.has_access(is_mod=True, from_lecture=True)
        )
    
    @staticmethod
    def __lambda_primary_join_public_publish_media():
        from .medium import PublishMedium
        return sql.and_(
            PublishMedium.lecture_id == Lecture.id,
            PublishMedium.has_access(is_mod=False, from_lecture=True)
        )
    
    @staticmethod
    def __lambda_primary_join_target_media():
        from .medium import TargetMedium
        return sql.and_(
            TargetMedium.lecture_id == Lecture.id,
            TargetMedium.is_active(usable_check=False, from_lecture=True)
        )
    
    @staticmethod
    def __lambda_primary_join_source_media():
        from .medium import SourceMedium
        return sql.and_(
            SourceMedium.lecture_id == Lecture.id,
            SourceMedium.is_active(from_lecture=True)
        )
    
    course: Mapped["Course"] = api_mapped(
        relationship(
            primaryjoin=lambda: Course.id == Lecture.course_id,
            back_populates="lectures",
            lazy="raise_on_sql"
        ),
        ApiMany2OneRelationshipField(
            include_in_data=True, data_foreign_in_context=True
        )
    )
    chapters: Mapped[list[Chapter]] = relationship(
        back_populates="lecture",
        primaryjoin=lambda: sql.and_(Chapter.lecture_id == Lecture.id, Chapter.has_access(is_mod=True)),
        order_by=Chapter.start_time.asc(),
        lazy="raise_on_sql"
    )
    # noinspection PyUnresolvedReferences
    publish_media: Mapped[list["PublishMedium"]] = relationship(
        back_populates="lecture",
        primaryjoin=__lambda_primary_join_publish_media,
        lazy="raise_on_sql"
    )
    public_chapters: Mapped[list[Chapter]] = relationship(
        back_populates="lecture",
        primaryjoin=lambda: sql.and_(Chapter.lecture_id == Lecture.id, Chapter.has_access(is_mod=False)),
        order_by=Chapter.start_time.asc(),
        lazy="raise_on_sql",
        viewonly=True
    )
    # noinspection PyUnresolvedReferences
    public_publish_media: Mapped[list["PublishMedium"]] = relationship(
        back_populates="lecture",
        primaryjoin=__lambda_primary_join_public_publish_media,
        lazy="raise_on_sql",
        viewonly=True
    )
    # noinspection PyUnresolvedReferences
    target_media: Mapped[list["TargetMedium"]] = relationship(
        back_populates="lecture",
        primaryjoin=__lambda_primary_join_target_media,
        lazy="raise_on_sql"
    )
    source_media: Mapped[list["SourceMedium"]] = relationship(
        back_populates="lecture",
        primaryjoin=__lambda_primary_join_source_media,
        lazy="raise_on_sql"
    )
    
    @property
    def effective_view_permissions(self) -> EffectiveViewPermissions:
        return self.view_permissions.get_effective(self.course.effective_view_permissions)
    
    @api_include_in_data(
        type_id="chapter[]",
        data_id="chapters",
        data_if=lambda lec, args: args.include_chapters,
        data_notes="Only present if requested. Sorted according to the start_time (ascending order)"
    )
    def _data_get_chapters(self, is_mod: bool):
        return self.chapters if is_mod else self.public_chapters
    
    @api_include_in_data(
        type_id="publish_medium[]",
        data_id="publish_media",
        data_if=lambda lec, args: args.include_media,
        data_notes="Only present if requested. All (public) media. May be empty"
    )
    def _data_get_publish_media(self, is_mod: bool):
        return self.publish_media if is_mod else self.public_publish_media
    
    @api_include_in_data(
        type_id="string[]",
        data_notes="possible values for string: `public` (Everyone has access, array will contain no other values), "
                   "`rwth`, `moodle`, `fsmpi`, `password`; Empty array indicates no access"
    )
    def authentication_methods(self):
        return self.effective_view_permissions.get_authentication_methods()
    
    @property
    def effective_media_process_obj(self) -> MediaProcess:
        if self.media_process is not None:
            return MediaProcess.from_json(CJsonValue(self.media_process).as_object())
        return MediaProcess.from_json(CJsonValue(self.course.media_process).as_object())
    
    @hybrid_method
    def is_active(self, from_course: bool = False, **kwargs):
        cond = super().is_active(**kwargs)
        if not from_course:
            cond &= self.hybrid_rel(self.course, "is_active", **kwargs)
        return cond
    
    @hybrid_method
    def has_access(self,
                   is_mod: bool,
                   visibility_check: bool = True,
                   from_course: bool = False,
                   **kwargs):
        cond = super().has_access(is_mod, visibility_check=visibility_check, from_course=from_course, **kwargs)
        if not from_course:
            cond &= self.hybrid_rel(self.course, "has_access", is_mod, visibility_check=visibility_check, **kwargs)
        return cond
    
    @classmethod
    def load_options(cls, is_mod: bool, to_load: list[RelationshipLoadMarker], from_course=False, **kwargs) -> list[ExecutableOption]:
        options = super().load_options(is_mod, to_load, **kwargs)
        
        if LOAD_LECTURE_CHAPTERS in to_load:
            options.append(
                orm.selectinload(Lecture.chapters if is_mod else Lecture.public_chapters).options(
                    # Causes no extra sql, just that the back reference is set (relationship has simple join)
                    orm.immediateload(Chapter.lecture)
                )
            )
        
        if LOAD_LECTURE_PUBLISH_MEDIA in to_load:
            from .medium import PublishMedium
            options.append(
                orm.selectinload(Lecture.publish_media if is_mod else Lecture.public_publish_media).options(
                    *PublishMedium.load_options(is_mod, to_load, from_lecture=True)
                )
            )
        
        if LOAD_LECTURE_TARGET_MEDIA in to_load:
            from .medium import TargetMedium
            options.append(
                orm.selectinload(Lecture.target_media).options(
                    *TargetMedium.load_options(is_mod, to_load)
                )
            )
        
        if LOAD_LECTURE_SOURCE_MEDIA in to_load:
            from .medium import SourceMedium
            options.append(
                orm.selectinload(Lecture.source_media).options(
                    *SourceMedium.load_options(is_mod, to_load, from_lecture=True)
                )
            )
        
        if from_course:
            # Causes no extra sql, just that the back reference is set (relationship has simple join)
            options.append(orm.immediateload(Lecture.course))
        elif LOAD_LECTURE_COURSE in to_load:
            options.append(orm.joinedload(Lecture.course).options(
                *Course.load_options(is_mod, to_load)
            ))
        
        return options


FORBIDDEN_COURSE_ID_STRING_VALUES = ["courses", "faq", "imprint", "licenses", "search", "site", "internal"]


class Course(DeletableApiObject, VisibilityApiObject, ApiViewPermissionsObject, Base):
    
    handle: Mapped[str] = api_mapped(
        mapped_column(String(32, STRING_COLLATION), nullable=False, index=True),
        ApiStringField(
            min_length=1, max_length=32, unique=True,
            regex=f"(?!(?:{'|'.join(FORBIDDEN_COURSE_ID_STRING_VALUES)})$)[a-zA-Z0-9_-]+",
            include_in_config=True,
            include_in_data=True,
            data_notes="Unique"
        )
    )
    full_name: Mapped[str] = api_mapped(
        mapped_column(Text(collation=STRING_COLLATION), nullable=False, default=""),
        ApiStringField(
            min_length=1, max_length=256,
            include_in_config=True, config_directly_modifiable=True,
            include_in_data=True
        )
    )
    short_name: Mapped[str] = api_mapped(
        mapped_column(String(32, STRING_COLLATION), nullable=False, default=""),
        ApiStringField(
            min_length=1, max_length=32,
            include_in_config=True, config_directly_modifiable=True,
            include_in_data=True
        )
    )
    semester: Mapped[str] = api_mapped(
        mapped_column(String(6, STRING_COLLATION), nullable=False, default=""),
        ApiStringField(
            max_length=6, regex="([0-9]{4}(ws|ss)|none)",
            include_in_config=True, config_directly_modifiable=True,
            include_in_data=True
        )
    )
    organizer: Mapped[str] = api_mapped(
        mapped_column(Text(collation=STRING_COLLATION), nullable=False, default=""),
        ApiStringField(
            max_length=256,
            include_in_config=True, config_directly_modifiable=True,
            include_in_data=True
        )
    )
    topic: Mapped[str] = api_mapped(
        mapped_column(String(32, STRING_COLLATION), nullable=False, default=""),
        ApiStringField(
            include_in_config=True, config_directly_modifiable=True,
            include_in_data=True
        )
    )
    description: Mapped[str] = api_mapped(
        mapped_column(Text(collation=STRING_COLLATION), nullable=False, default=""),
        ApiStringField(
            max_length=8192,
            include_in_config=True, config_directly_modifiable=True,
            include_in_data=True
        )
    )
    show_chapters_on_course: Mapped[bool] = api_mapped(
        mapped_column(nullable=False, default=False),
        ApiBooleanField(
            include_in_config=True,
            include_in_data=True,
            data_notes="If true, chapters should be shown on the course page"
        )
    )
    authentication_information: Mapped[str] = api_mapped(
        mapped_column(Text(collation=STRING_COLLATION), nullable=False, default=""),
        ApiStringField(
            max_length=8192,
            include_in_config=True,
            include_in_data=True,
            data_notes="May contain additional information on how the user can authenticate themselves"
        )
    )
    listed: Mapped[bool] = api_mapped(
        mapped_column(nullable=False, default=True),
        ApiBooleanField(
            include_in_config=True,
            include_in_data=True, data_only_mod=True,
            data_notes="If false, course can only be visited with its url (Unless user is moderator)"
        )
    )
    internal_comment: Mapped[str] = api_mapped(
        mapped_column(Text(collation=STRING_COLLATION), nullable=False, default=""),
        ApiStringField(
            max_length=8192,
            include_in_config=True, config_directly_modifiable=True,
            include_in_data=True, data_only_mod=True
        )
    )
    
    allow_download: Mapped[bool] = api_mapped(
        mapped_column(nullable=False, default=True),
        ApiBooleanField(
            include_in_config=True
        )
    )
    allow_embed: Mapped[bool] = api_mapped(
        mapped_column(nullable=False, default=False),
        ApiBooleanField(
            include_in_config=True,
            include_in_data=True
        )
    )
    media_process: Mapped[dict] = api_mapped(
        mapped_column(postgresql.JSONB, nullable=False, default=MediaProcess([], [], True).to_json()),
        ApiMediaProcessField(
            include_in_config=True
        )
    )
    disable_automatic_media_process_scheduler: Mapped[bool] = api_mapped(
        mapped_column(nullable=False, default=False),
        ApiBooleanField(
            include_in_config=True,
            # Lecture also has this field. Processor only works automatically if both fields are False
        )
    )
    
    # noinspection PyUnresolvedReferences
    responsible_users: Mapped[list["User"]] = api_mapped(
        relationship(
            secondary=responsible_table,
            back_populates="responsible_courses"
        ),
        ApiAny2ManyRelationshipField(
            include_in_config=True
        )
    )
    lectures: Mapped[list[Lecture]] = relationship(
        back_populates="course",
        primaryjoin=lambda: sql.and_(Lecture.course_id == Course.id, Lecture.has_access(is_mod=True, from_course=True)),
        order_by=Lecture.time.asc(),
        lazy="raise_on_sql"
    )
    public_lectures: Mapped[list[Lecture]] = relationship(
        back_populates="course",
        primaryjoin=lambda: sql.and_(Lecture.course_id == Course.id, Lecture.has_access(is_mod=False, from_course=True)),
        order_by=Lecture.time.asc(),
        lazy="raise_on_sql",
        viewonly=True
    )
    
    @property
    def effective_view_permissions(self):
        return self.view_permissions.get_effective(DEFAULT_VIEW_PERMISSIONS)
    
    @api_include_in_data(
        type_id="string[]",
        data_notes="Default authentication methods for lectures in this course. Some Lectures might have different "
                   "methods (A lecture object always contains its authentication methods). See lecture for possible values"
    )
    def default_authentication_methods(self):
        return self.effective_view_permissions.get_authentication_methods()
    
    @api_include_in_data(
        type_id="lecture[]",
        data_id="lectures",
        data_if=lambda course, args: args.include_lectures,
        data_notes="Only present if requested. All (public) lectures. Sorted by their time ascending. Lectures will "
                   "include chapters and publish media"
    )
    def _data_lectures(self, is_mod: bool):
        return self.lectures if is_mod else self.public_lectures
    
    @hybrid_method
    def has_access(self,
                   is_mod: bool,
                   visibility_check: bool = True,
                   ignore_unlisted: bool = False,
                   **kwargs):
        cond = super().has_access(is_mod, visibility_check, **kwargs)
        if not is_mod and visibility_check and not ignore_unlisted:
            cond &= self.listed
        return cond
    
    @classmethod
    def load_options(cls, is_mod: bool, to_load: list[RelationshipLoadMarker], **kwargs) -> list[ExecutableOption]:
        options = super().load_options(is_mod, to_load, **kwargs)
        
        if LOAD_COURSE_LECTURES in to_load:
            options.append(
                orm.selectinload(Course.lectures if is_mod else Course.public_lectures).options(
                    *Lecture.load_options(is_mod, to_load, from_course=True)
                )
            )
        return options
