from datetime import datetime
from enum import Enum

from videoag_common.miscellaneous import *
from videoag_common.database import *
from videoag_common.api_object import *
from .user import User


class JobState(Enum):
    READY = "ready"
    SPAWNING = "spawning"  # Job is in K8s but wasn't scheduled yet
    RUNNING = "running"  # Job is running in K8s
    FINISHED = "finished"
    FINISHED_AND_PROCESSED = "finished_and_processed"
    FAILED = "failed"
    FAILED_AND_PROCESSED = "failed_and_processed"
    CANCELED = "canceled"


_JOB_STATE_ENUM = create_enum_type(JobState)


class Job(ApiObject, Base):
    ERROR_CODE_UNKNOWN_TYPE = "unknown_type"
    ERROR_CODE_SPAWNING_FAILURE = "spawning_failure"
    
    __table_args__ = (
        CheckConstraint(
            "(on_end_event_type IS NULL) = (on_end_event_data IS NULL)",
            name="check_event",
            comment="Event type and data must be both set or both null"
        ),
        CheckConstraint(
            "cause_job_id IS NULL OR cause_user_id IS NULL",
            name="check_only_one_cause"
        )
    )

    status: Mapped[JobState] = api_mapped(
        mapped_column(_JOB_STATE_ENUM, nullable=False, index=True, default=JobState.READY),
        ApiEnumField(
            include_in_data=True
        )
    )
    type: Mapped[str] = api_mapped(
        mapped_column(String(collation=STRING_COLLATION), nullable=False),
        ApiStringField(
            include_in_data=True
        )
    )
    input_data: Mapped[JsonTypes] = api_mapped(
        mapped_column(postgresql.JSONB, nullable=False),
        ApiJsonField(
            include_in_data=True
        )
    )
    output_data: Mapped[JsonTypes] = api_mapped(
        mapped_column(postgresql.JSONB, nullable=True),
        ApiJsonField(
            include_in_data=True
        )
    )
    
    on_end_event_type: Mapped[str] = api_mapped(
        mapped_column(String(collation=STRING_COLLATION), nullable=True),
        ApiStringField(
            include_in_data=True
        )
    )
    on_end_event_data: Mapped[JsonTypes] = api_mapped(
        mapped_column(postgresql.JSONB, nullable=True),
        ApiJsonField(
            include_in_data=True
        )
    )
    creation_time: Mapped[datetime] = api_mapped(
        mapped_column(sql.DateTime(), nullable=False, server_default=sql.text("CURRENT_TIMESTAMP")),
        ApiDatetimeField(
            include_in_data=True
        )
    )
    run_start_time: Mapped[datetime] = api_mapped(
        mapped_column(sql.DateTime(), nullable=True),
        ApiDatetimeField(
            include_in_data=True
        )
    )
    run_end_time: Mapped[datetime] = api_mapped(
        mapped_column(sql.DateTime(), nullable=True),
        ApiDatetimeField(
            include_in_data=True
        )
    )
    cause_job_id: Mapped[int] = mapped_column(ForeignKey("job.id"), nullable=True, index=True)
    cause_user_id: Mapped[int] = mapped_column(ForeignKey("user.id"), nullable=True, index=True)
    
    
    # TODO monitoring???
    # TODO progress percent
    
    cause_job: Mapped["Job"] = api_mapped(
        relationship(
            primaryjoin=lambda: Job.id == Job.cause_job_id,
            lazy="select"  # Lazy loading because this is a recursive relationship
        ),
        ApiMany2OneRelationshipField(
            may_be_none=True,  # TODO because this is from Job to Job it can't figure out if its nullable
            include_in_data=True, data_foreign_in_context=True
        )
    )
    cause_user: Mapped[User] = api_mapped(
        relationship(
            primaryjoin=lambda: User.id == Job.cause_user_id,
            lazy="select"  # Lazy loading because this is a recursive relationship
        ),
        ApiMany2OneRelationshipField(
            may_be_none=True,
            include_in_data=True, data_foreign_in_context=True
        )
    )
    
    @staticmethod
    def create_event(event_type: str, event_data: JsonTypes, cause_job_id: int or None = None):
        return Job(
            type="handle_event",
            input_data={
                "type": event_type,
                "data": event_data
            },
            cause_job_id=cause_job_id
        )
    
    def set_error(self, error_code: str, error_message: str or None = None):
        self.status = JobState.FAILED
        self.output_data = {
            "error_code": error_code,
            "error_message": error_message
        }
