from enum import StrEnum
from typing import TypeVar, Generic

from videoag_common.miscellaneous import *
from videoag_common.database import *
from videoag_common.api_object import *


class ViewPermissionsType(StrEnum):
    PUBLIC = "public"
    PRIVATE = "private"
    AUTHENTICATION = "authentication"
    INHERIT = "inherit"


VIEW_PERMISSIONS_TYPE_ENUM = create_enum_type(ViewPermissionsType)


class ViewPermissions:
    
    def __init__(self,
                 type: ViewPermissionsType,
                 rwth_authentication: bool = False,
                 fsmpi_authentication: bool = False,
                 moodle_course_ids: list[int] or None = None,
                 passwords: dict[str, str] or None = None):
        super().__init__()
        self.type = type
        self.rwth_authentication = rwth_authentication
        self.fsmpi_authentication = fsmpi_authentication
        self.moodle_course_ids = moodle_course_ids
        self.passwords = passwords
        
        is_auth = type == ViewPermissionsType.AUTHENTICATION
        if (rwth_authentication and not is_auth) \
                or (fsmpi_authentication and not is_auth) \
                or ((moodle_course_ids is not None) != is_auth) \
                or ((passwords is not None) != is_auth) \
                or rwth_authentication is None \
                or fsmpi_authentication is None:
            raise TypeError("Invalid view permissions")
        if is_auth \
                and not rwth_authentication \
                and not fsmpi_authentication \
                and len(moodle_course_ids) == 0 \
                and len(passwords) == 0:
            raise RuntimeError("At least one authentication method must be present for type AUTHENTICATION")
    
    def get_effective(self, parent_permissions: "EffectiveViewPermissions"):
        return EffectiveViewPermissions(
            self,
            parent_permissions
        )
    
    def to_json(self) -> dict:
        if self.type == ViewPermissionsType.AUTHENTICATION:
            return {
                "type": self.type.value,
                "rwth_authentication": self.rwth_authentication,
                "fsmpi_authentication": self.fsmpi_authentication,
                "moodle_course_ids": self.moodle_course_ids,
                "passwords": self.passwords
            }
        else:
            return {
                "type": self.type.value
            }
    
    def __eq__(self, __value):
        return (
            isinstance(__value, ViewPermissions)
            and self.type == __value.type
            and self.rwth_authentication == __value.rwth_authentication
            and self.fsmpi_authentication == __value.fsmpi_authentication
            and self.passwords == __value.passwords
            and (True if self.type != ViewPermissionsType.AUTHENTICATION
                 else set(self.moodle_course_ids) == set(__value.moodle_course_ids))
        )


class EffectiveViewPermissions(ViewPermissions):
    
    def __init__(self,
                 object_permissions: ViewPermissions,
                 inherited_permissions: ViewPermissions = ViewPermissions(ViewPermissionsType.PUBLIC)):
        """
        :param object_permissions: The permissions set for the object
        :param inherited_permissions: Only relevant if `permissions.type == INHERIT`. May not have type INHERIT
        """
        effective = object_permissions
        if object_permissions.type == ViewPermissionsType.INHERIT:
            if inherited_permissions.type == ViewPermissionsType.INHERIT:
                raise RuntimeError("No indirect inheritance allowed")
            effective = inherited_permissions
        super().__init__(
            effective.type,
            effective.rwth_authentication,
            effective.fsmpi_authentication,
            effective.moodle_course_ids,
            effective.passwords
        )
        self.object_permissions = object_permissions
    
    def get_authentication_methods(self) -> list[str]:
        match self.type:
            case ViewPermissionsType.PUBLIC:
                return ["public"]
            case ViewPermissionsType.PRIVATE:
                return []
            case ViewPermissionsType.AUTHENTICATION:
                auth_list_json: list[str] = []
                if self.rwth_authentication:
                    auth_list_json.append("rwth")
                if self.fsmpi_authentication:
                    auth_list_json.append("fsmpi")
                if len(self.moodle_course_ids) > 0:
                    auth_list_json.append("moodle")
                if len(self.passwords) > 0:
                    auth_list_json.append("password")
                return auth_list_json
            case ViewPermissionsType.INHERIT:
                raise ValueError("Got INHERIT permission from get_effective_view_permissions")
            case _:
                raise ValueError(f"Unknown type {self.type}")


DEFAULT_VIEW_PERMISSIONS = ViewPermissions(ViewPermissionsType.PUBLIC)


def view_permissions_from_json(json: CJsonValue or JsonTypes) -> "ViewPermissions":
    is_client = isinstance(json, CJsonValue)
    if not is_client:
        json = CJsonValue(json)
    json = json.as_object()
    try:
        type_str: str = json.get_string("type", 100)
        try:
            type = ViewPermissionsType(type_str)
        except ValueError:
            raise json.raise_error(f"Unknown type '{truncate_string(type_str)}'")
        rwth_authentication = False
        fsmpi_authentication = False
        moodle_course_ids = None
        passwords = None
        if type == ViewPermissionsType.AUTHENTICATION:
            rwth_authentication = json.get_bool("rwth_authentication")
            fsmpi_authentication = json.get_bool("fsmpi_authentication")
            moodle_course_ids = [v.as_sint32() for v in json.get_array("moodle_course_ids")]
            passwords_json = json.get_object("passwords")
            passwords = {k: passwords_json.get_string(k, 1024) for k in passwords_json.keys()}
            
            if not rwth_authentication \
                    and not fsmpi_authentication \
                    and len(moodle_course_ids) == 0 \
                    and len(passwords) == 0:
                raise ApiClientException(ERROR_BAD_REQUEST(f"For type {ViewPermissionsType.AUTHENTICATION.value} at"
                                                           f" least one authentication method must be present"))
        else:
            json.raise_if_present("rwth_authentication")
            json.raise_if_present("fsmpi_authentication")
            json.raise_if_present("moodle_course_ids")
            json.raise_if_present("passwords")
        return ViewPermissions(
            type,
            rwth_authentication,
            fsmpi_authentication,
            moodle_course_ids,
            passwords
        )
    except ApiClientException as e:
        if is_client:
            raise e
        raise RuntimeError(f"Invalid json: {e.error.message}")


_O = TypeVar("_O", bound="ApiViewPermissionsObject")


class ApiViewPermissionsField(ApiConfigField[_O], Generic[_O]):
    
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
    
    def config_get_type_id(self) -> str:
        return "view_permissions"
    
    def post_init_check(self, context: FieldContext):
        super().post_init_check(context)
    
    def config_get_value(self, object_db: _O) -> JsonTypes:
        return object_db.view_permissions.to_json()
    
    @staticmethod
    def _set_perms_to_object(object_db: _O, perms: ViewPermissions):
        object_db.view_perm_type = perms.type
        object_db.view_perm_rwth_auth = perms.rwth_authentication
        object_db.view_perm_fsmpi_auth = perms.fsmpi_authentication
        if perms.type == ViewPermissionsType.AUTHENTICATION:
            object_db.view_perm_moodle_course_ids = perms.moodle_course_ids
            object_db.view_perm_passwords = perms.passwords
        else:
            object_db.view_perm_moodle_course_ids = None
            object_db.view_perm_passwords = None
    
    def config_set_value(self, session: SessionDb, object_db: _O, new_client_value: CJsonValue):
        ApiViewPermissionsField._set_perms_to_object(object_db, view_permissions_from_json(new_client_value))
        pass
    
    def config_get_default(self) -> JsonTypes:
        return DEFAULT_VIEW_PERMISSIONS.to_json()


class ApiViewPermissionsObject(ApiObject):
    __api_fields__ = [
        ApiViewPermissionsField(
            include_in_config=True, config_id="view_permissions"
        )
    ]
    view_perm_type: Mapped[ViewPermissionsType] = mapped_column(VIEW_PERMISSIONS_TYPE_ENUM, nullable=False, default=ViewPermissionsType.INHERIT)
    view_perm_rwth_auth: Mapped[bool] = mapped_column(nullable=False, default=False)
    view_perm_fsmpi_auth: Mapped[bool] = mapped_column(nullable=False, default=False)
    view_perm_moodle_course_ids: Mapped[list[int]] = mapped_column(sql.types.ARRAY(sql.types.Integer), nullable=True, default=None)
    view_perm_passwords: Mapped[dict] = mapped_column(sql.types.JSON(), nullable=True, default=None)
    
    @property
    def view_permissions(self):
        is_auth = self.view_perm_type == ViewPermissionsType.AUTHENTICATION
        return ViewPermissions(
            self.view_perm_type,
            self.view_perm_rwth_auth,
            self.view_perm_fsmpi_auth,
            self.view_perm_moodle_course_ids if is_auth else None,
            self.view_perm_passwords if is_auth else None,
        )
