import json
import os
from unittest import TestCase
from pathlib import Path

from api.database import *


# noinspection PyProtectedMember
def reset_database():
    data_script = Path(os.path.join(os.getcwd(), api.config['DB_DATA'])).read_text(encoding="UTF-8")
    if api.config["DATABASE"]["engine"] == "postgres":
        import re
        # Postgres sequences get out of sync when inserting rows with id
        data_script = re.sub(
            "--\\s*\\$POSTGRES_FIX_SEQUENCE\\s*\\(\\s*([a-zA-Z0-9_-]+)\\s*,\\s*([a-zA-Z0-9_-]+)\\s*\\)",
            """\
SELECT setval(pg_get_serial_sequence('\\1', '\\2'), COALESCE((SELECT MAX("\\2")+1 FROM "\\1"), 1), false);\
""", data_script)
    
    def _trans(session: SessionDb):
        session.execute(sql.text(data_script))
        session.commit()
    database.execute_write_transaction(_trans)


class DatabaseTest(TestCase):
    
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
    
    def setUp(self):
        super().setUp()
        reset_database()
    
    @staticmethod
    def _get_index_of_object(list: list, object):
        i: int = 0
        for entry in list:
            if entry is object:
                return i
            i += 1
        raise ValueError("Object not found")
    
    @staticmethod
    def _list_minus(first_list: list, second_list: list):
        res = []
        for l1 in first_list:
            for l2 in second_list:
                if l1 == l2:
                    break
            else:  # Loop did not break
                res.append(l1)
        return res
    
    @staticmethod
    def _json_check_is_ignored(ignored_keys: list[str] or None, path: str):
        for ignore_key in ignored_keys:
            if (ignore_key == path) if ignore_key.count(".") != 0 else (ignore_key == path.split(".")[-1]):
                return True
        return False
    
    def assertJsonEqual(self,
                        expected: object,
                        data: object,
                        ignore_array_order_keys: list[str] or None = None,
                        ignore_value_keys: list[str] or None = None,
                        path: str = "$",
                        only_print_errors: bool = False) -> int:
        """
        Checks if two json values are equal. All differences are printed and then an exception is raised
        (unless only_print_errors is True)
        :param expected: Expected data.
        :param data: Data to check
        :param ignore_array_order_keys: A list of keys which refer to arrays for which the order of elements does not
                                        matter. Format is either a full path or a single key. A full path are the object
                                        keys separated by '.' and arrays with '[<index>]'. The beginning is a '$'.
                                        Example: '$.values[0].type' refers to value 'type' in the first object of array
                                        'values'.
                                        A single key is one without any '.' This matches all arrays which are saved in
                                        an object with that key.
                                        This does NOT work nested (e.g. an array which may be in any order cannot
                                        contain arrays which itself may be in any order). ignore_value_keys only works
                                        for single key entries.
        :param ignore_value_keys: Same format as ignore_array_order_keys. All values wich are saved with a matching key
                                  are ignored (May exists, May not exists, May have a different value than in
                                  expected).
        :param path: The path where expected is
        :param only_print_errors: If true no exception will be raised
        :return: The amount of errors/differences found
        """
        if ignore_array_order_keys is None:
            ignore_array_order_keys = []
        if ignore_value_keys is None:
            ignore_value_keys = []
        if DatabaseTest._json_check_is_ignored(ignore_value_keys, path):
            return 0
        if type(data) != type(expected):
            msg = f"{path}: Expected type {type(expected)} got {type(data)} instead"
            if only_print_errors:
                print(msg)
                return 1
            raise AssertionError(msg)
        error_count = 0
        if isinstance(expected, dict):
            assert isinstance(data, dict)
            
            missing_keys = []
            unexpected_keys = []
            for key in set(expected.keys()).union(set(data.keys())):
                sub_path = f"{path}.{key}"
                if DatabaseTest._json_check_is_ignored(ignore_value_keys, sub_path):
                    continue
                if key in expected:
                    if key not in data:
                        missing_keys.append(key)
                        continue
                    error_count += self.assertJsonEqual(expected[key], data[key],
                                                        ignore_array_order_keys,
                                                        ignore_value_keys,
                                                        sub_path,
                                                        True)
                else:
                    assert key in data
                    unexpected_keys.append(key)
            
            if len(missing_keys) > 0 or len(unexpected_keys) > 0:
                error_count += 1
                print(f"{path}: Missing key(s): {missing_keys} Unexpected key(s): {unexpected_keys}")
        elif isinstance(expected, list):
            assert isinstance(data, list)
            if len(data) != len(expected):
                error_count += 1
                print(f"{path} Expected {len(expected)} value(s) but got {len(data)}")
            else:
                ignore_order = DatabaseTest._json_check_is_ignored(ignore_array_order_keys, path)
                
                if ignore_order:
                    expected_filtered = DatabaseTest._remove_values_with_keys(expected, ignore_value_keys)
                    data_filtered = DatabaseTest._remove_values_with_keys(data, ignore_value_keys)
                    missing_values = DatabaseTest._list_minus(expected_filtered, data_filtered)
                    unexpected_values = DatabaseTest._list_minus(data_filtered, expected_filtered)
                    assert len(missing_values) == len(unexpected_values)
                    
                    for expected_value, unexpected_value in zip(missing_values, unexpected_values):
                        error_count += self.assertJsonEqual(expected_value, unexpected_value, ignore_array_order_keys,
                                                            ignore_value_keys,
                                                            f"{path}[?]",
                                                            True)
                else:
                    i: int = 0
                    for expected_value, value in zip(expected, data):
                        error_count += self.assertJsonEqual(expected_value, value, ignore_array_order_keys,
                                                            ignore_value_keys,
                                                            f"{path}[{i}]", True)
                        i += 1
        else:
            try:
                self.assertEqual(expected, data, f"At {path}")
            except Exception as e:
                print(e)
                error_count += 1
        if error_count > 0 and not only_print_errors:
            raise AssertionError(f"Got {error_count} error(s)")
        return error_count
    
    @staticmethod
    def _remove_values_with_keys(object: dict or list, keys_to_remove: list[str]):
        if isinstance(object, dict):
            new_object = {}
            for key, value in object.items():
                if key in keys_to_remove:
                    continue
                if isinstance(value, (dict, list)):
                    value = DatabaseTest._remove_values_with_keys(value, keys_to_remove)
                new_object[key] = value
            return new_object
        elif isinstance(object, list):
            new_object = []
            for value in object:
                if isinstance(value, (dict, list)):
                    value = DatabaseTest._remove_values_with_keys(value, keys_to_remove)
                new_object.append(value)
            return new_object
        else:
            raise TypeError(f"Got {object} but expected a dict or list")
    
    # noinspection PyMethodMayBeStatic
    def assert_contains(self, collection, obj: object):
        if obj not in collection:
            raise AssertionError(f"{obj} not in {collection}")
